--------------------------------------------
-- Company:        Liverpool University 
-- Engineer:       Ilya Tsurin
-- 
-- Create Date:    21:41:25 11/19/2009 
-- Design Name:    Neutron detection
-- Module Name:    Aboard - Behavioral 
-- Project Name:   
-- Target Devices: XC3S500E-4FG320
-- Tool versions:  ISE 11.3 WebPack
-- Description: 
--
-- Dependencies:   Top Module
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments: 

-- FE MAC address convention:
--              S2:XX:XX:XX:XX:XX (hex)
-- S  = SSSS       (bin) - MC selection switch
-- 2  = 00LU       (bin) - IEEE convention: locally (L=1)
--                         administered unicast (U=0)
-- X - any arbitrary value (could be packet ID)
--------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


library UNISIM;
use UNISIM.VComponents.all;

entity aboard is
    Port ( Clock  : in std_logic;
           OSC_40 : in std_logic;

           ETH_CRS : in std_logic;
           ETH_COL : in std_logic;

           ETH_RX_DV  : in std_logic;
           ETH_RX_CLK : in std_logic;
			  ETH_RX_ERR : in std_logic;
           ETH_RXD    : in std_logic_vector (3 downto 0);

           ETH_TX_EN  : out std_logic;
           ETH_TX_CLK : in  std_logic;
           ETH_TX_ERR : out std_logic;
           ETH_TXD    : out std_logic_vector (3 downto 0);

			  ETH_MDC  : out   std_logic;
			  ETH_MDIO : inout std_logic;

           USB_nTXE : out   std_logic;
           USB_nRXF : out   std_logic;
           USB_nRD  : in    std_logic;
           USB_WR   : in    std_logic;
           USB_DAT  : inout std_logic_vector (7 downto 0);

           PMT  : in std_logic_vector (1 downto 0);
           SIT  : in std_logic_vector (3 downto 0);
           SITR : in std_logic;

           SYS_ADR : out std_logic_vector (3 downto 0);
-- SYS_ADR(3) = LED(4)
-- SYS_ADR(2) = LED(5)
-- SYS_ADR(1) = LED(6)
-- SYS_ADR(0) = LED(7)

           SYS_CLK : out std_logic_vector (3 downto 0);
-- SYS_CLK(1) = LED(0)
-- SYS_CLK(0) = LED(3)

           SYS_TRG : out std_logic_vector (3 downto 0);
-- SYS_TRG(0) = LED(2)

           SYS_RES : out std_logic_vector (3 downto 0);
-- SYS_RES(0) = LED(1)

           TST_SYN : out   std_logic;
           TST_TRG : out   std_logic;
           TST_DV0 : in    std_logic;
           TST_PUL : out   std_logic;
           TST_RES : out   std_logic;
           TST_SCL : out   std_logic;
           TST_SDA : inout std_logic;

	   SW   : in std_logic_vector(3 downto 0);
			  
	   BTN  : in std_logic;
	   BTS  : in std_logic;
	   BTW  : in std_logic;
	   BTE  : in std_logic;
	   BTC  : in std_logic;

	   ROTA : in std_logic;
	   ROTB : in std_logic;

		CLK_50MHZ:    in std_logic;
		CLK_AUX:      in std_logic;
		CLK_SMA:      in std_logic;

		AD_CONV:     out std_logic;

		AMP_CS:      out std_logic;
		AMP_DOUT:     in std_logic;
		AMP_SHDN:    out std_logic;

		DAC_CLR:     out std_logic;
		DAC_CS:      out std_logic;

		DS_WIRE:   inout std_logic;

		FPGA_M0:   inout std_logic;
		FPGA_M1:   inout std_logic;
		FPGA_M2:      in std_logic;
		FPGA_INIT_B: out std_logic;

-- ghost signals
--		FPGA_RDWR_B: out std_logic;
--		FPGA_HSWAP:   in std_logic;

		LCD_E:  out std_logic;
		LCD_RS: out std_logic;
		LCD_RW: out std_logic;

		PS2_CLK:     out std_logic;
		PS2_DATA:    out std_logic;

		RS232_DCE_RXD:  in std_logic;
		RS232_DCE_TXD: out std_logic;
		RS232_DTE_RXD:  in std_logic;
		RS232_DTE_TXD: out std_logic;
	
		SD_A:      out std_logic_vector(12 downto 0);
		SD_BA:     out std_logic_vector(1 downto 0);
		SD_CAS:    out std_logic;
		SD_CK_N:   out std_logic;
		SD_CK_P:   out std_logic;
		SD_CKE:    out std_logic;
		SD_CS:     out std_logic;
		SD_DQ:   inout std_logic_vector(15 downto 0);
		SD_LDM:    out std_logic;
		SD_LDQS: inout std_logic;
		SD_RAS:    out std_logic;
		SD_UDM:    out std_logic;
		SD_UDQS: inout std_logic;
		SD_WE:     out std_logic;
		SD_CK_FB:   in std_logic;

-- SF_A(24) is shared with TST_RES
		SF_A1:   inout std_logic_vector(23 downto 20);
		SF_A0:     out std_logic_vector(19 downto 0);

		SF_BYTE:   out std_logic;
		SF_CE0:    out std_logic;
		SF_D:    inout std_logic_vector(15 downto 1);
		SF_OE:     out std_logic;
		SF_STS:     in std_logic;
		SF_WE:     out std_logic;

		SPI_MISO:   in std_logic;
		SPI_MOSI:  out std_logic;
		SPI_SCK: inout std_logic;
		SPI_SS_B:  out std_logic;
		SPI_ALT_CS_JP11: out std_logic; 

		VGA_BLUE:  out std_logic;
		VGA_GREEN: out std_logic;
		VGA_HSYNC: out std_logic;
		VGA_RED:   out std_logic;
		VGA_VSYNC: out std_logic

-- ghost signals
--		XC_CMD:    out std_logic_vector(1 downto 0);
--		XC_CPLD_EN: in std_logic;
--		XC_D:    inout std_logic_vector(2 downto 0);
--		XC_TRIG:    in std_logic;
--		XC_GCK0:   out std_logic;
--		GCLK10:    out std_logic
);

	attribute pinnum: string;

	attribute pinnum of Clock:      signal is "C9";
	attribute pinnum of OSC_40:     signal is "E10";
	
	attribute pinnum of ETH_CRS:    signal is "U13";
	attribute pinnum of ETH_COL:    signal is "U6";

	attribute pinnum of ETH_RX_DV:  signal is "V2";
	attribute pinnum of ETH_RX_CLK: signal is "V3";
	attribute pinnum of ETH_RX_ERR: signal is "U14";
	attribute pinnum of ETH_RXD:    signal is "V14,U11,V8,U14";

	attribute pinnum of ETH_TX_EN:  signal is "P15";
	attribute pinnum of ETH_TX_CLK: signal is "T7";
	attribute pinnum of ETH_TX_ERR: signal is "R6";
	attribute pinnum of ETH_TXD:    signal is "T5,R5,T15,R11";

	attribute pinnum of ETH_MDC:  signal is "P9";
	attribute pinnum of ETH_MDIO: signal is "U5";	
	
	attribute pinnum of USB_nTXE:   signal is "E8";
	attribute pinnum of USB_nRXF:   signal is "B4";
	attribute pinnum of USB_nRD:    signal is "C7";
	attribute pinnum of USB_WR:     signal is "F8";
	attribute pinnum of USB_DAT:    signal is "B6,F7,E7,D5,D7,C5,A6,A4";

	attribute pinnum of PMT:  signal is "C12,D12";
	attribute pinnum of SIT:  signal is "A15,B15,C3,C15";
	attribute pinnum of SITR: signal is "D9";

	attribute pinnum of SYS_ADR: signal is "C11,D11,E9,F9";
	attribute pinnum of SYS_CLK: signal is "B16,C14,B13,E12";
	attribute pinnum of SYS_TRG: signal is "A16,B14,A13,E11";
	attribute pinnum of SYS_RES: signal is "D14,A14,F12,F11";

	attribute pinnum of TST_SYN: signal is "D10";
	attribute pinnum of TST_TRG: signal is "G9";
	attribute pinnum of TST_DV0: signal is "B11";
	attribute pinnum of TST_PUL: signal is "A8";
	attribute pinnum of TST_RES: signal is "A11";
	attribute pinnum of TST_SCL: signal is "E13";
	attribute pinnum of TST_SDA: signal is "C4";

	attribute pinnum of SW: signal is "N17,H18,L14,L13";

	attribute pinnum of BTN: signal is "V4";
	attribute pinnum of BTS: signal is "K17";
	attribute pinnum of BTW: signal is "D18";
	attribute pinnum of BTE: signal is "H13";
	attribute pinnum of BTC: signal is "V16";

	attribute pinnum of ROTA: signal is "K18";
	attribute pinnum of ROTB: signal is "G18";

end aboard;

architecture Behavioral of aboard is

-- Clock management
--------------------------------------------
	signal CLK: std_logic;	-- Buffered ETH_TX_CLK (25 MHz)
	signal SCK: std_logic;	-- 50 MHz derived from ETH_TX_CLK
--------------------------------------------

-- System wake up
--------------------------------------------
	signal PLL_REG: std_logic_vector(3 downto 0);
	signal PLL_DEL: std_logic_vector(3 downto 0);
	signal RX_SOF : std_logic;								-- RX (receive) frame
	signal DV_DEL : std_logic;
--------------------------------------------

-- Ethernet input package
--------------------------------------------
	signal RXB_CLK: std_logic;
	signal RXB_CNT: natural range 0 to 2047;
	signal RXB_DAT: std_logic_vector(7 downto 0);	-- merging input nibbles

	signal MAC_CHK: std_logic_vector(7 downto 0);
	signal UNICAST: std_logic;								-- MAC match, chip select

	signal FE_ADR:  std_logic_vector(47 downto 0);

--		FE_ADR(47 downto 44) switch
--		FE_ADR(43 downto 40) unicast
--		FE_ADR(39 downto 0)  packet ID

	signal HOST_ADR: std_logic_vector(47 downto 0);
--------------------------------------------

-- Command flags for a single-loop processor
--------------------------------------------
	signal CTR_REG: std_logic_vector(2 downto 0);
	signal SPI_REG: std_logic_vector(2 downto 0);
	signal THR_REG: std_logic_vector(13 downto 0);

	signal DO_THR: std_logic;
	signal DO_SPI: std_logic;
	signal DO_ADC: std_logic;

	signal Peak0: std_logic;
	signal Peak1: std_logic;

	signal SIGN: std_logic;
	signal TRIG: std_logic;
	
	signal TRG_LAT: std_logic;
	signal TRG_DEL: std_logic;

	signal Sample:  natural := 10;
	signal SAM_CNT: natural range 0 to 15;
--------------------------------------------

-- ADC cycle
--------------------------------------------
	signal ADC_GO: std_logic;
	signal BIT_CNT: natural range 0 to 63;

	signal ADC0: std_logic_vector(13 downto 0);
	signal ADC1: std_logic_vector(13 downto 0);

	signal ADC_RG0: std_logic_vector(13 downto 0);
	signal ADC_RG1: std_logic_vector(13 downto 0);
--------------------------------------------

-- Data RAM controls
--------------------------------------------
	signal RAM_GTE: std_logic_vector(0 downto 0);
	signal RAM_CLK: std_logic;

	signal RAM_ADR: natural range 0 to 2047;
	signal RAM_WAD: std_logic_vector(10 downto 0);
	signal RAM_RAD: std_logic_vector(10 downto 0);

	signal INP_DAT: std_logic_vector(7 downto 0);
	signal OUT_DAT: std_logic_vector(7 downto 0);
--------------------------------------------

-- Ethernet packet pacer
--------------------------------------------
	signal CYC_CNT: natural range 0 to 2097151;
	signal CYC_PER: std_logic;

	signal EVT_CNT: natural range 0 to 2147483647;
	signal EVT_NUM: std_logic_vector(31 downto 0);
--------------------------------------------

-- Ethernet packet length
--------------------------------------------
	signal TX_CYCLE: std_logic;
	signal TXB_CNT: natural range 0 to 2047;
--------------------------------------------

-- Ethernet packet attributes
--------------------------------------------
	signal DAT_CNT: natural range 0 to 2047;
	signal DAT_LIM: natural range 0 to 2047;
	signal DAT_LEN: std_logic_vector(10 downto 0);
--------------------------------------------

-- Checksum variables
--------------------------------------------
	signal CRC_CLK: std_logic;
	signal CRC_INI: std_logic;
	signal CRC_DAT: std_logic_vector(7 downto 0);
	signal CRC_CRC: std_logic_vector(31 downto 0);
	signal CRC_RES: std_logic_vector(31 downto 0);
--------------------------------------------

-- Beetle control
--------------------------------------------
	signal TST_DAT: std_logic;
	signal TST_DEN: std_logic;
--------------------------------------------

-- declaring PLL component
--------------------------------------------
component pll
	port (CLKIN_IN:   in std_logic;
			CLK0_OUT:  out std_logic;
			CLK2X_OUT: out std_logic);
end component;
--------------------------------------------

-- declaring CRC component
--------------------------------------------
	component CRC
		port (
			  Clock: in std_logic;
			  Init : in std_logic;
			  Data : in std_logic_vector(7 downto 0);
			  CRC  : out std_logic_vector(31 downto 0)
			  );
	end component;
--------------------------------------------

-- declaring RAM component
--------------------------------------------
	component RAM
		port (
			clka:   in std_logic;
			wea:    in std_logic_vector(0 downto 0);
			addra:  in std_logic_vector(10 downto 0);
			dina:   in std_logic_vector(7 downto 0);
			clkb:   in std_logic;
			addrb:  in std_logic_vector(10 downto 0);
			doutb: out std_logic_vector(7 downto 0)
				);
	end component;
--------------------------------------------

-- declaring comparator component
--------------------------------------------
	component COMP
		port (
				a: in std_logic_vector(13 downto 0);
				b: in std_logic_vector(13 downto 0);
				a_gt_b: out std_logic
				);
	end component;
--------------------------------------------

-- OK, here the entity starts


begin

-- Permanent assignments
--------------------------------------------

-- Resolve pin conflicts
--------------------------------------------
	AMP_SHDN <= '0';

	DAC_CLR  <= '0';
	DAC_CS   <= '1';
	DS_WIRE  <= 'Z';

-- M2..M0 are selected by jumpers
	FPGA_M0  <= 'Z';
	FPGA_M1  <= 'Z';
--	FPGA_M2  <= 'Z';
	FPGA_INIT_B <= '1';

	LCD_E    <= '0';
	LCD_RS   <= '0';
	LCD_RW   <= '0';

	PS2_CLK  <= '0';
	PS2_DATA <= '0';

	RS232_DCE_TXD <= '0';
	RS232_DTE_TXD <= '0';

	SD_A     <= (others => '0');
	SD_BA    <= "00";
	SD_CAS   <= '1'; -- NOP
	SD_CK_N  <= '1';
	SD_CK_P  <= '0';
	SD_CKE   <= '0'; -- disable clock
	SD_CS    <= '1'; -- disable command decoder
	SD_DQ    <= (others => 'Z');
	SD_LDM   <= '0';
	SD_LDQS  <= 'Z';
	SD_RAS   <= '1'; -- NOP
	SD_UDM   <= '0';
	SD_UDQS  <= 'Z';
	SD_WE    <= '1'; -- NOP

	SF_A1    <= (others => 'Z');
	SF_A0    <= (others => '0');
	SF_BYTE  <= '0';
	SF_CE0   <= '1';
	SF_D     <= (others => 'Z');
	SF_OE    <= '1';
	SF_WE    <= '1';

	SPI_SS_B <= '1';
	SPI_ALT_CS_JP11 <= 'Z';

	VGA_BLUE  <='0';
	VGA_GREEN <='0';
	VGA_HSYNC <='0';
	VGA_RED   <='0';
	VGA_VSYNC <='0';
--------------------------------------------

-- FE clock fanout
--------------------------------------------
	SYS_ADR <= (others => 'Z');
--	SYS_CLK <= (others => 'Z');
	
	SYS_CLK(3) <= TRG_LAT;
	SYS_CLK(2) <= TRG_DEL;
	SYS_CLK(1) <= '1' when RAM_GTE = "1" else '0';
	SYS_CLK(0) <= RAM_CLK;

	SYS_TRG <= (others => 'Z');
	SYS_RES <= (others => 'Z');
--------------------------------------------

-- Test block is not yet used
--------------------------------------------
	TST_TRG <= '0';
	TST_SYN <= '0';
	TST_RES <= '0';
	TST_PUL <= '0';
--------------------------------------------

-- Ethernet collision is not used
--------------------------------------------
	ETH_TX_ERR <= '0';
	ETH_MDIO   <= 'Z';
	ETH_MDC    <= '0';
--------------------------------------------

-- FE data bus is not used
--------------------------------------------
	USB_nTXE <= 'Z';
	USB_nRXF <= 'Z';
	USB_DAT  <= (others => 'Z');
--------------------------------------------

-- Frequency  multiplier
--------------------------------------------
U0: pll
		port map (
			CLKIN_IN     => ETH_TX_CLK,
			CLK0_OUT     => CLK,
			CLK2X_OUT    => SCK);
--------------------------------------------

-- !!! Wake up !!!
-- Capture the frame delimiter (Result CLK <-)
--------------------------------------------
	process(ETH_RX_CLK, ETH_RXD)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			PLL_DEL <= ETH_RXD;
		end if;
	end process;
--------------	
	process(ETH_RX_CLK, PLL_DEL)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '0') then
			PLL_REG <= PLL_DEL;
		end if;
	end process;
--------------------------------------------

-- Start of the RX frame (CLK ->)
--------------------------------------------
	process(ETH_RX_CLK, ETH_RX_DV, ETH_RXD, PLL_REG)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			if (ETH_RX_DV = '1') then
				if (ETH_RXD = x"D" and PLL_REG = x"5") then
					RX_SOF <= '1';
				else
					null;
				end if;
			else
				RX_SOF <= '0';
			end if;
		end if;
	end process;
--------------------------------------------

-- Signal for the clock synchronisation (CLK ->)
--------------------------------------------
	process(ETH_RX_CLK, ETH_RX_DV)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			DV_DEL <= ETH_RX_DV;
		end if;
	end process;
--------------------------------------------

-- RX nibble swapper and data clock (CLK <-)
-- (synchronize phase with RX_SOF signal)
--------------------------------------------
	process(ETH_RX_CLK, DV_DEL, RX_SOF)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '0') then
			if (DV_DEL = '1' and RX_SOF = '0') then
				RXB_CLK <= '0';
			else
				RXB_CLK <= not RXB_CLK;
			end if;
		end if;
	end process;
--------------------------------------------

-- RX byte counter (CLK ->)
--------------------------------------------
	process(RXB_CLK, RX_SOF)
	begin
		if (RXB_CLK'event and RXB_CLK = '1') then
			if (RX_SOF = '1') then
				RXB_CNT <= RXB_CNT + 1;
			else
				RXB_CNT <= 0;
			end if;
		end if;
	end process;
--------------------------------------------

-- Compile bytes from nibbles. Data are
-- ready for latching by (CLK ->) under
-- NIR_SWP = '0' condition
--------------------------------------------
	process(ETH_RX_CLK, RXB_CLK, ETH_RXD)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			if (RXB_CLK = '1') then
				RXB_DAT(3 downto 0) <= ETH_RXD;
			else
				null;
			end if;
		end if;
	end process;
--------------
	RXB_DAT(7 downto 4) <= ETH_RXD;
--------------------------------------------

-- Validate FE addressing (unicast)
--------------------------------------------
	process(ETH_RX_CLK, RXB_CNT, RXB_CLK, RXB_DAT)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			if (RXB_CNT = 1 and RXB_CLK = '0') then
				MAC_CHK <= RXB_DAT;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- Locally administered unicast
--------------------------------------------
	process(MAC_CHK)
	begin
		if (MAC_CHK(1 downto 0) = "10") then
			UNICAST <= '1';
		else
			UNICAST <= '0';
		end if;
	end process;
--------------------------------------------

-- Capture the RX data
--------------------------------------------
	process(ETH_RX_CLK, UNICAST, RXB_CLK, RXB_CNT,
			  MAC_CHK, RXB_DAT, DO_THR, DO_SPI)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			if (UNICAST = '1' and RXB_CLK = '0') then
				case RXB_CNT is
					when 2 =>
						FE_ADR(47 downto 40)   <= MAC_CHK;
						FE_ADR(39 downto 32)   <= RXB_DAT;
					when 3 =>
						FE_ADR(31 downto 24)   <= RXB_DAT;
					when 4 =>
						FE_ADR(23 downto 16)   <= RXB_DAT;
					when 5 =>
						FE_ADR(15 downto 8)    <= RXB_DAT;
					when 6 =>
						FE_ADR(7 downto 0)     <= RXB_DAT;
					when 7 =>
						HOST_ADR(47 downto 40) <= RXB_DAT;
					when 8 =>
						HOST_ADR(39 downto 32) <= RXB_DAT;
					when 9 =>
						HOST_ADR(31 downto 24) <= RXB_DAT;
					when 10 =>
						HOST_ADR(23 downto 16) <= RXB_DAT;
					when 11 =>
						HOST_ADR(15 downto 8)  <= RXB_DAT;
					when 12 =>
						HOST_ADR(7 downto 0)   <= RXB_DAT;

-- skip Dsize (2 bytes)
 
					when 15 =>
						   CTR_REG <= RXB_DAT(2 downto 0);
					when 16 =>
						if    (DO_THR = '1') then
							THR_REG(13 downto 8) <= RXB_DAT(5 downto 0);
						elsif (DO_SPI = '1') then
							SPI_REG <= RXB_DAT(2 downto 0);
						else
							null;
						end if;
					when 17 =>
						if (DO_THR = '1') then
							THR_REG(7 downto 0) <= RXB_DAT;
						else
							null;
						end if;
					when others =>
						null;
				end case;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- Hardware access
--------------------------------------------
	process(CTR_REG)
	begin
		case CTR_REG(1 downto 0) is
			when "01" =>
				DO_THR <= '1';
				DO_SPI <= '0';
				DO_ADC <= '0';				
			when "10" =>
				DO_THR <= '0';
				DO_SPI <= '1';
				DO_ADC <= '0';
			when "11" =>
				DO_THR <= '0';
				DO_SPI <= '0';
				DO_ADC <= '1';
			when others =>
				DO_THR <= '0';
				DO_SPI <= '0';
				DO_ADC <= '0';
		end case;
	end process;
--------------
	process(DO_SPI, SPI_REG, DO_ADC, SCK)
	begin
		if (DO_SPI = '1') then
			AMP_CS   <= SPI_REG(2);
			SPI_SCK  <= SPI_REG(1);
			SPI_MOSI <= SPI_REG(0);
		else
			AMP_CS   <= '1';
			SPI_SCK  <= DO_ADC and SCK;
			SPI_MOSI <= '0';
		end if;
	end process;
--------------
	SIGN <= CTR_REG(2);
--------------------------------------------

-- Bit counter @ 50 MHz
--------------------------------------------
	process(SCK, ADC_GO)
	begin
		if (SCK'event and SCK = '1') then
			if (ADC_GO = '1') then
				BIT_CNT <= 0;
			else
				BIT_CNT <= BIT_CNT + 1;
			end if;
		end if;
	end process;
--------------------------------------------

-- ADC sampling strobe
--------------------------------------------
	process(SCK, BIT_CNT)
	begin
		if (SCK'event and SCK = '0') then
			case BIT_CNT is
				when 0 =>
					ADC_GO <= '0';
				when 33 =>
					ADC_GO <= '1';
				when others =>
					null;
			end case;
		end if;
	end process;
--------------
	AD_CONV <= DO_ADC and ADC_GO;
--------------------------------------------

-- Compile data
--------------------------------------------
	process(SCK, BIT_CNT, SPI_MISO)
	begin
		if (SCK'event and SCK = '0') then
			case BIT_CNT is
				when 2 =>
					ADC0(13) <= not SPI_MISO;
				when 3 =>
					ADC0(12) <= SPI_MISO;
				when 4 =>
					ADC0(11) <= SPI_MISO;
				when 5 =>
					ADC0(10) <= SPI_MISO;
				when 6 =>
					ADC0(9) <= SPI_MISO;
				when 7 =>
					ADC0(8) <= SPI_MISO;
				when 8 =>
					ADC0(7) <= SPI_MISO;
				when 9 =>
					ADC0(6) <= SPI_MISO;
				when 10 =>
					ADC0(5) <= SPI_MISO;
				when 11 =>
					ADC0(4) <= SPI_MISO;
				when 12 =>
					ADC0(3) <= SPI_MISO;
				when 13 =>
					ADC0(2) <= SPI_MISO;
				when 14 =>
					ADC0(1) <= SPI_MISO;
				when 15 =>
					ADC0(0) <= SPI_MISO;
				when 18 =>
					ADC1(13) <= not SPI_MISO;
				when 19 =>
					ADC1(12) <= SPI_MISO;
				when 20 =>
					ADC1(11) <= SPI_MISO;
				when 21 =>
					ADC1(10) <= SPI_MISO;
				when 22 =>
					ADC1(9) <= SPI_MISO;
				when 23 =>
					ADC1(8) <= SPI_MISO;
				when 24 =>
					ADC1(7) <= SPI_MISO;
				when 25 =>
					ADC1(6) <= SPI_MISO;
				when 26 =>
					ADC1(5) <= SPI_MISO;
				when 27 =>
					ADC1(4) <= SPI_MISO;
				when 28 =>
					ADC1(3) <= SPI_MISO;
				when 29 =>
					ADC1(2) <= SPI_MISO;
				when 30 =>
					ADC1(1) <= SPI_MISO;
				when 31 =>
					ADC1(0) <= SPI_MISO;
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- Buffer ADC data
--------------------------------------------
	process(ADC_GO, ADC0, ADC1)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			ADC_RG0 <= ADC0;
			ADC_RG1 <= ADC1;
		end if;
	end process;
--------------------------------------------

-- Threshold (unsigned vectors)
--------------------------------------------
U1: COMP
		port map (
			a      => ADC_RG0,
			b      => THR_REG,
			a_gt_b => Peak0);
--------------
U2: COMP
		port map (
			a      => ADC_RG1,
			b      => THR_REG,
			a_gt_b => Peak1);
--------------------------------------------

-- Trigger condition
--------------------------------------------
	process(SIGN, Peak0, Peak1)
	begin
		if (SIGN = '1') then
			TRIG <= Peak0 or Peak1;
		else
			TRIG <= not (Peak0 and Peak1);
		end if;
	end process;
--------------------------------------------

-- Trigger pulse
--------------------------------------------
	process(ADC_GO, TRIG)
	begin
		if (ADC_GO'event and ADC_GO = '0') then
			TRG_LAT <= TRIG;
		end if;
	end process;
--------------------------------------------

-- Trigger latency
--------------------------------------------
	process(ADC_GO, TRG_LAT)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			TRG_DEL <= TRG_LAT;
		end if;
	end process;
--------------------------------------------

-- RAM write enable
--------------------------------------------
	process(ADC_GO, SAM_CNT, Sample, CYC_CNT, TRIG, TRG_DEL, DAT_CNT)
	begin
		if (ADC_GO'event and ADC_GO = '0') then
			if (SAM_CNT = Sample) then
				RAM_GTE <= "0";				
			elsif ((CYC_CNT <= (1470407-Sample)) and
					((TRIG = '1') and (TRG_DEL = '0')) and
					 (DAT_CNT <= 1459)) then
				RAM_GTE <= "1";
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- Counter for number of samples
--------------------------------------------
	process(ADC_GO, RAM_GTE)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			if (RAM_GTE = "1") then
				SAM_CNT <= SAM_CNT + 1;
			else
				SAM_CNT <= 0;
			end if;
		end if;
	end process;
--------------------------------------------

-- Data size counter
--------------------------------------------
	process(ADC_GO, CYC_PER, RAM_GTE)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			if (CYC_PER = '1') then
				DAT_CNT <= 0;
			elsif (RAM_GTE = "1") then
				DAT_CNT <= DAT_CNT + 4;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- RAM write strobe
--------------------------------------------
	process(RAM_GTE, DAT_CNT, SCK, BIT_CNT)
	begin
		if ((RAM_GTE = "0") or (1500 <= DAT_CNT)) then
			RAM_CLK <= '0';
		elsif (SCK'event and SCK = '0') then
			case BIT_CNT is
				when 4 =>
					RAM_CLK <= '1';
				when 8 =>
					RAM_CLK <= '0';
				when 12 =>
					RAM_CLK <= '1';
				when 16 =>
					RAM_CLK <= '0';
				when 20 =>
					RAM_CLK <= '1';
				when 24 =>
					RAM_CLK <= '0';
				when 28 =>
					RAM_CLK <= '1';
				when 32 =>
					RAM_CLK <= '0';
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- Interleave RAM data
--------------------------------------------
	process(SCK, BIT_CNT, ADC_RG0, ADC_RG1)
	begin
		if (SCK'event and SCK = '0') then
			case BIT_CNT is
				when 0 =>
					INP_DAT(7) <= '1';
					INP_DAT(6 downto 0) <= ADC_RG0(13 downto 7);
				when 8 =>
					INP_DAT(7) <= '0';
					INP_DAT(6 downto 0) <= ADC_RG0(6 downto 0);
				when 16 =>
					INP_DAT(7) <= '0';
					INP_DAT(6 downto 0) <= ADC_RG1(13 downto 7);
				when 24 =>
					INP_DAT(7) <= '0';
					INP_DAT(6 downto 0) <= ADC_RG1(6 downto 0);
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- RAM write address
--------------------------------------------
	process(CYC_PER, RAM_CLK)
	begin
		if (CYC_PER = '1') then
			RAM_ADR <= 0;
		elsif (RAM_CLK'event and RAM_CLK = '0') then
			RAM_ADR <= RAM_ADR + 1;
		end if;
	end process;
--------------
	RAM_WAD <= conv_std_logic_vector(RAM_ADR, 11);
--------------------------------------------

-- Period for data transmission
--------------------------------------------
	process(ADC_GO, CYC_PER)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			if (CYC_PER = '0') then
				CYC_CNT <= CYC_CNT + 1;
			else
				CYC_CNT <= 0;
			end if;
		end if;
	end process;
--------------
	process(ADC_GO, CYC_CNT)
	begin
		if (ADC_GO'event and ADC_GO = '0') then
			case CYC_CNT is
				when 0 =>
					CYC_PER <= '0';
				when 1470588 =>
					CYC_PER <= '1';
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- Ethernet packet number
--------------------------------------------
	process(ADC_GO, CYC_PER)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			if (CYC_PER = '1') then
				EVT_CNT <= EVT_CNT + 1;
			end if;
		end if;
	end process;
--------------
	EVT_NUM <= conv_std_logic_vector(EVT_CNT, 32);
--------------------------------------------

-- Transmission cycle
--------------------------------------------
	process(ADC_GO, CYC_CNT)
	begin
		if (ADC_GO'event and ADC_GO = '0') then
			case CYC_CNT is
				when 1470408 =>
					TX_CYCLE <= '1';
				when 1470588 =>
					TX_CYCLE <= '0';
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- Counter limits
--------------------------------------------
	process(DAT_CNT)
	begin
		if (DAT_CNT <= 46) then
			DAT_LIM <= 46;
		elsif (1500 <= DAT_CNT) then
			DAT_LIM <= 1500;
		else
			DAT_LIM <= DAT_CNT;
		end if;
	end process;
--------------
	DAT_LEN <= conv_std_logic_vector(DAT_LIM, 11);
--------------------------------------------

-- TX nibble swapper = CRC32 clock (CLK ->)
--------------------------------------------
	process(ETH_TX_CLK)
	begin
		if (ETH_TX_CLK'event and ETH_TX_CLK = '1') then
			CRC_CLK <= not CRC_CLK;
		end if;
	end process;
--------------------------------------------

-- TX byte counter (CLK ->) @ 12.5 MHz for
-- sending out 1 ethernet packet (1500 bytes)
-- per second for up to 375 Hz event trigger
-- rate (event size 4 bytes)
--------------------------------------------
	process(CRC_CLK, TX_CYCLE)
	begin
		if (CRC_CLK'event and CRC_CLK = '1') then
			if (TX_CYCLE = '1') then
				TXB_CNT <= TXB_CNT + 1;
			else
				TXB_CNT <= 0;
			end if;
		end if;
	end process;
--------------------------------------------

-- Drive ETH_TX_EN signal
-- Readout is postponed if it conflicts
-- with logging of zero-suppressed data 
--------------------------------------------
	process(ETH_TX_CLK, CRC_CLK, TXB_CNT, DAT_LIM)
	begin
		if (ETH_TX_CLK'event and ETH_TX_CLK = '0') then
			if (CRC_CLK = '1') then

-- Causes J/K/ = x55 for the receiver on the other end
				if (TXB_CNT = 1) then
					ETH_TX_EN <= '1';
				elsif (TXB_CNT = (22 + DAT_LIM + 4 + 1)) then
					ETH_TX_EN <= '0';
				else
					null;
				end if;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- Initialize CRC variables (CLK <-)
--------------------------------------------
	process(ETH_TX_CLK, TXB_CNT, CRC_CLK)
	begin
		if (ETH_TX_CLK'event and ETH_TX_CLK = '0') then
			if (TXB_CNT = 9) then
				if (CRC_CLK = '1') then
					CRC_INI <= '1';
				else
					CRC_INI <= '0';
				end if;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- compile the Ethernet output packet
--------------------------------------------
	process(ETH_TX_CLK, TXB_CNT, CRC_CLK,
			  HOST_ADR, FE_ADR, EVT_NUM,
--			  DAT_LEN, OUT_DAT, CRC_CRC, CRC_RES)
			  RAM_WAD, OUT_DAT, CRC_CRC, CRC_RES)
	begin
		if (ETH_TX_CLK'event and ETH_TX_CLK = '0') then

-- 6 times preamble 0x55
--------------------------------------------
			if ((1 <= TXB_CNT) and (TXB_CNT <= 7)) then
				ETH_TXD <= x"5";

-- Delimiter 0xD5
--------------------------------------------

			elsif (TXB_CNT = 8) then
				if (CRC_CLK = '1') then
					ETH_TXD <= x"5";
				else
					ETH_TXD <= x"D";
				end if;

-- Destination (Host) MAC address
--------------------------------------------
			elsif (TXB_CNT = 9) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(43 downto 40);
				else
					ETH_TXD <= HOST_ADR(47 downto 44);
				end if;
			elsif (TXB_CNT = 10) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(35 downto 32);
				else
					ETH_TXD <= HOST_ADR(39 downto 36);
				end if;
			elsif (TXB_CNT = 11) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(27 downto 24);
				else
					ETH_TXD <= HOST_ADR(31 downto 28);
				end if;
			elsif (TXB_CNT = 12) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(19 downto 16);
				else
					ETH_TXD <= HOST_ADR(23 downto 20);
				end if;
			elsif (TXB_CNT = 13) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(11 downto 8);
				else
					ETH_TXD <= HOST_ADR(15 downto 12);
				end if;
			elsif (TXB_CNT = 14) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(3 downto 0);
				else
					ETH_TXD <= HOST_ADR(7 downto 4);
				end if;

-- Source (Front-end) address
--------------------------------------------
			elsif (TXB_CNT = 15) then
				if (CRC_CLK = '1') then
					ETH_TXD <= FE_ADR(43 downto 40);
				else
					ETH_TXD <= FE_ADR(47 downto 44);
				end if;
			elsif (TXB_CNT = 16) then
				if (CRC_CLK = '1') then
					ETH_TXD <= FE_ADR(35 downto 32);
				else
					ETH_TXD <= FE_ADR(39 downto 36);
				end if;
			elsif (TXB_CNT = 17) then
				if (CRC_CLK = '1') then
					ETH_TXD <= EVT_NUM(27 downto 24);
				else
					ETH_TXD <= EVT_NUM(31 downto 28);
				end if;
			elsif (TXB_CNT = 18) then
				if (CRC_CLK = '1') then
					ETH_TXD <= EVT_NUM(19 downto 16);
				else
					ETH_TXD <= EVT_NUM(23 downto 20);
				end if;
			elsif (TXB_CNT = 19) then
				if (CRC_CLK = '1') then
					ETH_TXD <= EVT_NUM(11 downto 8);
				else
					ETH_TXD <= EVT_NUM(15 downto 12);
				end if;
			elsif (TXB_CNT = 20) then
				if (CRC_CLK = '1') then
					ETH_TXD <= EVT_NUM(3 downto 0);
				else
					ETH_TXD <= EVT_NUM(7 downto 4);
				end if;

-- Packet length
--------------------------------------------
			elsif (TXB_CNT = 21) then
				if (CRC_CLK = '1') then
					ETH_TXD(2 downto 0) <= RAM_WAD(10 downto 8); -- DAT_LEN
					ETH_TXD(3) <= '0';
				else
					ETH_TXD <= (others => '0');
				end if;
			elsif (TXB_CNT = 22) then
				if (CRC_CLK = '1') then
					ETH_TXD <= RAM_WAD(3 downto 0); -- DAT_LEN
				else
					ETH_TXD <= RAM_WAD(7 downto 4); -- DAT_LEN
				end if;

-- CRC32
--------------------------------------------
			elsif (TXB_CNT = (22 + DAT_LIM + 1)) then
-- *** Important *** Use CRC_CRC Nibble here
				if (CRC_CLK = '1') then
					ETH_TXD <= CRC_CRC(3 downto 0);
				else
					ETH_TXD <= CRC_CRC(7 downto 4);
				end if;
			elsif (TXB_CNT = (22 + DAT_LIM + 2)) then
				if (CRC_CLK = '1') then
					ETH_TXD <= CRC_RES(11 downto 8);
				else
					ETH_TXD <= CRC_RES(15 downto 12);
				end if;
			elsif (TXB_CNT = (22 + DAT_LIM + 3)) then
				if (CRC_CLK = '1') then
					ETH_TXD <= CRC_RES(19 downto 16);
				else
					ETH_TXD <= CRC_RES(23 downto 20);
				end if;
			elsif (TXB_CNT = (22 + DAT_LIM + 4)) then
				if (CRC_CLK = '1') then
					ETH_TXD <= CRC_RES(27 downto 24);
				else
					ETH_TXD <= CRC_RES(31 downto 28);
				end if;

-- Data block
--------------------------------------------
			else
				if (CRC_CLK = '1') then
					ETH_TXD <= OUT_DAT(3 downto 0);
				else
					ETH_TXD <= OUT_DAT(7 downto 4);
				end if;
			end if;
		end if;
	end process;
--------------------------------------------

-- CRC data
--------------------------------------------
	process(TXB_CNT, HOST_ADR, FE_ADR,
--			  EVT_NUM, DAT_LEN, OUT_DAT)
			  EVT_NUM, RAM_WAD, OUT_DAT)
	begin

-- Destination (Host) MAC address
--------------------------------------------
		if (TXB_CNT = 9) then
			CRC_DAT <= HOST_ADR(47 downto 40);
		elsif (TXB_CNT = 10) then
			CRC_DAT <= HOST_ADR(39 downto 32);
		elsif (TXB_CNT = 11) then
			CRC_DAT <= HOST_ADR(31 downto 24);
		elsif (TXB_CNT = 12) then
			CRC_DAT <= HOST_ADR(23 downto 16);
		elsif (TXB_CNT = 13) then
			CRC_DAT <= HOST_ADR(15 downto 8);
		elsif (TXB_CNT = 14) then
			CRC_DAT <= HOST_ADR(7 downto 0);

-- Source (Front-end) address
--------------------------------------------
		elsif (TXB_CNT = 15) then
			CRC_DAT <= FE_ADR(47 downto 40);
		elsif (TXB_CNT = 16) then
			CRC_DAT <= FE_ADR(39 downto 32);
		elsif (TXB_CNT = 17) then
			CRC_DAT <= EVT_NUM(31 downto 24);
		elsif (TXB_CNT = 18) then
			CRC_DAT <= EVT_NUM(23 downto 16);
		elsif (TXB_CNT = 19) then
			CRC_DAT <= EVT_NUM(15 downto 8);
		elsif (TXB_CNT = 20) then
			CRC_DAT <= EVT_NUM(7 downto 0);

-- Packet length
--------------------------------------------
		elsif (TXB_CNT = 21) then
			CRC_DAT(7 downto 3) <= "00000";
			CRC_DAT(2 downto 0) <= RAM_WAD(10 downto 8); -- DAT_LEN
		elsif (TXB_CNT = 22) then
			CRC_DAT <= RAM_WAD(7 downto 0); -- DAT_LEN

-- Data block
-------------------------------------------
		else
			CRC_DAT <= OUT_DAT;
		end if;
	end process;
--------------------------------------------
--------------------------------------------

-- Calculate CRC32 for transmitted data
-------------------------------------------
	U3:CRC
	port map (
		Clock => CRC_CLK,
		Init  =>	CRC_INI,
		Data  =>	CRC_DAT,
		CRC   =>	CRC_CRC);
--------------------------------------------

-- Keep the CRC result (the last data byte)
--------------------------------------------
	process(CRC_CLK, TXB_CNT, DAT_LIM, CRC_CRC)
	begin
		if (CRC_CLK'event and CRC_CLK = '0') then
			if (TXB_CNT = (22 + DAT_LIM + 1)) then
				CRC_RES <= CRC_CRC;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------
--------------------------------------------

-- RAM read address
--------------------------------------------
	process(CRC_CLK, TXB_CNT)
	begin
		if (CRC_CLK'event and CRC_CLK = '0') then
			RAM_RAD <= conv_std_logic_vector(TXB_CNT, 11);
		end if;
	end process;
--------------------------------------------

-- RAM for the event data
--------------------------------------------
	U5:RAM
	port map(
		clka   => RAM_CLK,
		wea    => RAM_GTE,
		addra  => RAM_WAD,
		dina   => INP_DAT,
		clkb   => CRC_CLK,
		addrb  => RAM_RAD,
		doutb  => OUT_DAT);
--------------------------------------------

--	OUT_DAT <= (others => '0');


-- Reserved for future use
--------------------------------------------
	TST_SYN <= '0';
	TST_TRG <= '0';
	TST_PUL <= '0';
	TST_RES <= '0';

	TST_SCL <= 'Z';
	TST_SDA <= 'Z';
--------------------------------------------

end Behavioral;
