/* Alibava control via USB FT245R or raw Ethernet interface  */
/* Copyright I.Tsurin Liverpool University V1.0 (30.09.2011) */
/*                                                           */
/* http://twiki.ific.uv.es/twiki/bin/view/Atlas/ALiBaVa      */                        
/* Acknowledgment: Paulo Pedreiras, Out.2005 RCAI 05/06      */
/* http://www.ieeta.pt/~pedreiras/resources/rcai/raw-sr.c    */

/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <ctype.h>

#include <signal.h>
#include <sys/ioctl.h>

#include <sys/socket.h>
#include <netinet/in.h>

#include <linux/if.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>  // the L2 protocols
#include <asm/types.h>


#define ETH_SAFE_LEN (ETH_DATA_LEN - 100)
#define DEVICE "eth0"
#define Dsize 46

int s = 0;		  // Socket descriptor
unsigned char *eth_head;  // byte pointer to ethenet header
unsigned char *eth_data;  // byte pointer to ethernet data field
struct ethhdr *eh;	  // structured pointer to ethernet header

struct sockaddr_ll socket_address;
void *buffer = NULL;

/* Local host NIC MAC address */
unsigned char  src_mac[6] = {0x00, 0x1D, 0x09, 0xB1, 0x46, 0x6F};

/* Front end address template */
unsigned char dest_mac[6] = {0x02, 0x00, 0x00, 0x00, 0x00, 0x00};

void sigint(int signum);
int eth_raw_init();

unsigned char l=1;
unsigned char O=0;


  FILE *fp;


/* Commands */
#define NOP   0x00
#define doTHR 0x01
#define doSPI 0x02
#define doADC 0x03

/* Gain codes */
#define G0   0x00
#define G1   0x01
#define G2   0x02
#define G5   0x03
#define G10  0x04
#define G20  0x05
#define G50  0x06
#define G100 0x07


/* SPI controls */
#define nCS0 O
#define nCS1 l

#define SCK0 O
#define SCK1 l

#define SDI0 O
#define SDI1 l

#define SPIoctet(nCS, SCK, Dbit) \
  ((nCS & 0x01)<<2 | (SCK & 0x01)<<1 | (Dbit & 0x01))

#define POS l
#define NEG O

/***************************************/
/*       Board controls                */
/***************************************/

void setTHR(int Value)
{
  int i;

  eth_data[0] = doTHR;
  eth_data[1] = (Value>>8)&0x3F;
  eth_data[2] =  Value&0xFF;
  i = sendto(s, buffer, (Dsize + ETHER_HDR_LEN), 0,
               ((struct sockaddr*)&socket_address), sizeof(socket_address));
  if (i == -1) { perror("sendto():"); exit(1); }
}


void SPIbit(unsigned char Dbit)
{
  int i;

  eth_data[0] = doSPI;
  eth_data[1] = SPIoctet(nCS0, SCK0, Dbit);
  i = sendto(s, buffer, (Dsize + ETHER_HDR_LEN), 0,
              ((struct sockaddr*)&socket_address), sizeof(socket_address));
  if (i == -1) { perror("sendto():"); exit(1); }

  eth_data[0] = doSPI;
  eth_data[1] = SPIoctet(nCS0, SCK1, Dbit);
  i = sendto(s, buffer, (Dsize + ETHER_HDR_LEN), 0,
              ((struct sockaddr*)&socket_address), sizeof(socket_address));
  if (i == -1) { perror("sendto():"); exit(1); }
}


void setSPI(unsigned char Gain0, unsigned char Gain1)
{
  int i;

  eth_data[0] = doSPI;
  eth_data[1] = SPIoctet(nCS1, SCK0, SDI0);
  i = sendto(s, buffer, (Dsize + ETHER_HDR_LEN), 0,
               ((struct sockaddr*)&socket_address), sizeof(socket_address));
  if (i == -1) { perror("sendto():"); exit(1); }

  eth_data[0] = doSPI;
  eth_data[1] = SPIoctet(nCS0, SCK0, SDI0);
  i = sendto(s, buffer, (Dsize + ETHER_HDR_LEN), 0,
               ((struct sockaddr*)&socket_address), sizeof(socket_address));
  if (i == -1) { perror("sendto():"); exit(1); }

  for (i=0; i<4; i++)
     {
       unsigned char Dbit = (Gain0>>(3-i)) & 0x01;
       SPIbit(Dbit);
     }

  for (i=0; i<4; i++)
     {
       unsigned char Dbit = (Gain1>>(3-i)) & 0x01;
       SPIbit(Dbit);
     }

  eth_data[0] = doSPI;
  eth_data[1] = SPIoctet(nCS1, SCK1, SDI0);
  i = sendto(s, buffer, (Dsize + ETHER_HDR_LEN), 0,
               ((struct sockaddr*)&socket_address), sizeof(socket_address));
  if (i == -1) { perror("sendto():"); exit(1); }
}


void RUNgo(unsigned char ACT)
{
  int i;

  eth_data[0] = (NEG<<2) | ACT;
  i = sendto(s, buffer, (Dsize + ETHER_HDR_LEN), 0,
               ((struct sockaddr*)&socket_address), sizeof(socket_address));
  if (i == -1) { perror("sendto():"); exit(1); }
}


/***************************************/
/*             main module             */
/*                                     */
/* send configuration to motherboard N */ 
/***************************************/

int main(int argc, char **argv)
{
  int i, ind = 0, *Wpt;

  int Nevents = 0;
  char *Fname = "./adata.bin";

  if ((argc==2) || (argc==3)) Nevents = atoi(argv[1]);
  if  (argc==3) Fname = argv[2];

/* Create socket */

  eth_raw_init();
  buffer = (void*)malloc(ETHER_MAX_LEN);

/* set structurised pointer to beginning of ethernet header */

  eth_head = buffer;
  eh = (struct ethhdr *)eth_head;

  memcpy((void *)eh->h_dest,  (void*)dest_mac, ETH_ALEN);
  memcpy((void *)eh->h_source,(void*)src_mac,  ETH_ALEN);

  eh->h_proto = htons(Dsize);
  eth_data = buffer + ETH_HLEN;


/* open data file */

  fp=fopen(Fname, "w+");

/* Configure boards */

  printf("Configure Aboard\n");

  setTHR(8190);
  setSPI(G1, G1);
  RUNgo(doADC);


/* Event loop */

  while(1)
    {
      unsigned char *Wpt;
      unsigned char Nevt[4];

/* Read NC buffer */

usleep(10);
      i = recvfrom(s, buffer, ETHER_MAX_LEN, 0, NULL, NULL);

      if (i == -1) { perror("recfrom():"); exit(1); }

      Wpt = buffer;

      for (i=0; i<18; i++) printf("%02X:", *Wpt++);
      printf("\n");

      ind++;
      if (ind == Nevents) break;

    } /* end while */

  RUNgo(NOP);

fclose(fp);
return 0;

} /* end main */



/***************************************/
/*            eth_raw_init             */
/***************************************/

int eth_raw_init()
{
  int i;
  struct ifreq ifr;
  int ifindex = 0;  // Ethernet Interface index
  
/* Open socket */

  s = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
  if (s == -1) { perror("socket():"); exit(1); }

  printf("Successfully opened socket: %i\n", s);
  
/*Get ethernet interface index */

  strncpy(ifr.ifr_name, DEVICE, IFNAMSIZ);
  if (ioctl(s, SIOCGIFINDEX, &ifr) == -1) { perror("SIOCGIFINDEX"); exit(1); }

  ifindex = ifr.ifr_ifindex;
  printf("Successfully got interface index: %i\n", ifindex);
  
/* retrieve corresponding MAC */

  if (ioctl(s, SIOCGIFHWADDR, &ifr) == -1) { perror("SIOCGIFINDEX"); exit(1); }

  for (i=0; i<6; i++) src_mac[i] = ifr.ifr_hwaddr.sa_data[i];

  printf("Host MAC address: ");
  for (i=0; i<6; i++) printf("%02X:", src_mac[i]);
  printf("\n");
  
/* prepare sockaddr_ll */

  socket_address.sll_family   = PF_PACKET;
  socket_address.sll_protocol = 0x00;
  socket_address.sll_ifindex  = ifindex;
  socket_address.sll_pkttype  = PACKET_BROADCAST;
  socket_address.sll_halen    = ETH_ALEN;
  socket_address.sll_addr[0]  = dest_mac[0];
  socket_address.sll_addr[1]  = dest_mac[1];
  socket_address.sll_addr[2]  = dest_mac[2];
  socket_address.sll_addr[3]  = dest_mac[3];
  socket_address.sll_addr[4]  = dest_mac[4];
  socket_address.sll_addr[5]  = dest_mac[5];
  socket_address.sll_addr[6]  = 0x00; 
  socket_address.sll_addr[7]  = 0x00;
														
  
/* establish signal handler */

  signal(SIGINT, sigint);
  printf("Established signal handler for SIGINT\n");
  
  return 0;
}

/***************************************/
/*          Socket Terminator          */
/***************************************/

void sigint(int signum)
{
  struct ifreq ifr;
	
  if (s == -1) return; // No socket has beeen created
	
  strncpy(ifr.ifr_name, DEVICE, IFNAMSIZ);
  ioctl(s, SIOCGIFFLAGS, &ifr);
  ifr.ifr_flags &= ~IFF_PROMISC;
  ioctl(s, SIOCSIFFLAGS, &ifr);
  close(s);

  free(buffer);
  fclose(fp);
	
  printf("Client terminating...\n");
  exit(0);
}

