--------------------------------------------
-- Company:        Liverpool University 
-- Engineer:       Ilya Tsurin
-- 
-- Create Date:    21:41:25 11/19/2009 
-- Design Name:    Neutron detection
-- Module Name:    Aboard - Behavioral 
-- Project Name:   
-- Target Devices: XC3S500E-4FG320
-- Tool versions:  ISE 11.3 WebPack
-- Description: 
--
-- Dependencies:   Top Module
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments: 

-- FE MAC address convention:
--              X2:XX:XX:XX:XX:XX (hex)
-- 2  = 00LU       (bin) - IEEE convention: locally (L=1)
--                         administered unicast (U=0)
-- X - any arbitrary value (could be packet ID)
--------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


library UNISIM;
use UNISIM.VComponents.all;

entity aboard is
    Port ( Clock  : in std_logic;
           OSC_40 : in std_logic;

           ETH_CRS : in std_logic;
           ETH_COL : in std_logic;

           ETH_RX_DV  : in std_logic;
           ETH_RX_CLK : in std_logic;
			  ETH_RX_ERR : in std_logic;
           ETH_RXD    : in std_logic_vector (3 downto 0);

           ETH_TX_EN  : out std_logic;
           ETH_TX_CLK : in  std_logic;
           ETH_TX_ERR : out std_logic;
           ETH_TXD    : out std_logic_vector (3 downto 0);

			  ETH_MDC  : out   std_logic;
			  ETH_MDIO : inout std_logic;

           USB_nTXE : out   std_logic;
           USB_nRXF : out   std_logic;
           USB_nRD  : in    std_logic;
           USB_WR   : in    std_logic;
           USB_DAT  : inout std_logic_vector (7 downto 0);

           PMT  : in std_logic_vector (1 downto 0);
           SIT  : in std_logic_vector (3 downto 0);
           SITR : in std_logic;

           SYS_ADR : out std_logic_vector (3 downto 0);
-- SYS_ADR(3) = LED(4)
-- SYS_ADR(2) = LED(5)
-- SYS_ADR(1) = LED(6)
-- SYS_ADR(0) = LED(7)

           SYS_CLK : out std_logic_vector (3 downto 0);
-- SYS_CLK(1) = LED(0)
-- SYS_CLK(0) = LED(3)

           SYS_TRG : out std_logic_vector (3 downto 0);
-- SYS_TRG(0) = LED(2)

           SYS_RES : out std_logic_vector (3 downto 0);
-- SYS_RES(0) = LED(1)

           TST_SYN : out   std_logic;
           TST_TRG : out   std_logic;
           TST_DV0 : in    std_logic;
           TST_PUL : out   std_logic;
           TST_RES : out   std_logic;
           TST_SCL : out   std_logic;
           TST_SDA : inout std_logic;

	   SW   : in std_logic_vector(3 downto 0);
			  
	   BTN  : in std_logic;
	   BTS  : in std_logic;
	   BTW  : in std_logic;
	   BTE  : in std_logic;
	   BTC  : in std_logic;

	   ROTA : in std_logic;
	   ROTB : in std_logic;

		CLK_50MHZ:    in std_logic;
		CLK_AUX:      in std_logic;
		CLK_SMA:      in std_logic;

		AD_CONV:     out std_logic;

		AMP_CS:      out std_logic;
		AMP_DOUT:     in std_logic;
		AMP_SHDN:    out std_logic;

		DAC_CLR:     out std_logic;
		DAC_CS:      out std_logic;

		DS_WIRE:   inout std_logic;

		FPGA_M0:   inout std_logic;
		FPGA_M1:   inout std_logic;
		FPGA_M2:      in std_logic;
		FPGA_INIT_B: out std_logic;

-- ghost signals
--		FPGA_RDWR_B: out std_logic;
--		FPGA_HSWAP:   in std_logic;

		LCD_E:  out std_logic;
		LCD_RS: out std_logic;
		LCD_RW: out std_logic;

		PS2_CLK:     out std_logic;
		PS2_DATA:    out std_logic;

		RS232_DCE_RXD:  in std_logic;
		RS232_DCE_TXD: out std_logic;
		RS232_DTE_RXD:  in std_logic;
		RS232_DTE_TXD: out std_logic;
	
		SD_A:      out std_logic_vector(12 downto 0);
		SD_BA:     out std_logic_vector(1 downto 0);
		SD_CAS:    out std_logic;
		SD_CK_N:   out std_logic;
		SD_CK_P:   out std_logic;
		SD_CKE:    out std_logic;
		SD_CS:     out std_logic;
		SD_DQ:   inout std_logic_vector(15 downto 0);
		SD_LDM:    out std_logic;
		SD_LDQS: inout std_logic;
		SD_RAS:    out std_logic;
		SD_UDM:    out std_logic;
		SD_UDQS: inout std_logic;
		SD_WE:     out std_logic;
		SD_CK_FB:   in std_logic;

-- SF_A(24) is shared with TST_RES
		SF_A1:   inout std_logic_vector(23 downto 20);
		SF_A0:     out std_logic_vector(19 downto 0);

		SF_BYTE:   out std_logic;
		SF_CE0:    out std_logic;
		SF_D:    inout std_logic_vector(15 downto 1);
		SF_OE:     out std_logic;
		SF_STS:     in std_logic;
		SF_WE:     out std_logic;

		SPI_MISO:   in std_logic;
		SPI_MOSI:  out std_logic;
		SPI_SCK: inout std_logic;
		SPI_SS_B:  out std_logic;
		SPI_ALT_CS_JP11: out std_logic; 

		VGA_BLUE:  out std_logic;
		VGA_GREEN: out std_logic;
		VGA_HSYNC: out std_logic;
		VGA_RED:   out std_logic;
		VGA_VSYNC: out std_logic

-- ghost signals
--		XC_CMD:    out std_logic_vector(1 downto 0);
--		XC_CPLD_EN: in std_logic;
--		XC_D:    inout std_logic_vector(2 downto 0);
--		XC_TRIG:    in std_logic;
--		XC_GCK0:   out std_logic;
--		GCLK10:    out std_logic
);

	attribute pinnum: string;

	attribute pinnum of Clock:      signal is "C9";
	attribute pinnum of OSC_40:     signal is "E10";
	
	attribute pinnum of ETH_CRS:    signal is "U13";
	attribute pinnum of ETH_COL:    signal is "U6";

	attribute pinnum of ETH_RX_DV:  signal is "V2";
	attribute pinnum of ETH_RX_CLK: signal is "V3";
	attribute pinnum of ETH_RX_ERR: signal is "U14";
	attribute pinnum of ETH_RXD:    signal is "V14,U11,V8,U14";

	attribute pinnum of ETH_TX_EN:  signal is "P15";
	attribute pinnum of ETH_TX_CLK: signal is "T7";
	attribute pinnum of ETH_TX_ERR: signal is "R6";
	attribute pinnum of ETH_TXD:    signal is "T5,R5,T15,R11";

	attribute pinnum of ETH_MDC:  signal is "P9";
	attribute pinnum of ETH_MDIO: signal is "U5";	
	
	attribute pinnum of USB_nTXE:   signal is "E8";
	attribute pinnum of USB_nRXF:   signal is "B4";
	attribute pinnum of USB_nRD:    signal is "C7";
	attribute pinnum of USB_WR:     signal is "F8";
	attribute pinnum of USB_DAT:    signal is "B6,F7,E7,D5,D7,C5,A6,A4";

	attribute pinnum of PMT:  signal is "C12,D12";
	attribute pinnum of SIT:  signal is "A15,B15,C3,C15";
	attribute pinnum of SITR: signal is "D9";

	attribute pinnum of SYS_ADR: signal is "C11,D11,E9,F9";
	attribute pinnum of SYS_CLK: signal is "B16,C14,B13,E12";
	attribute pinnum of SYS_TRG: signal is "A16,B14,A13,E11";
	attribute pinnum of SYS_RES: signal is "D14,A14,F12,F11";

	attribute pinnum of TST_SYN: signal is "D10";
	attribute pinnum of TST_TRG: signal is "G9";
	attribute pinnum of TST_DV0: signal is "B11";
	attribute pinnum of TST_PUL: signal is "A8";
	attribute pinnum of TST_RES: signal is "A11";
	attribute pinnum of TST_SCL: signal is "E13";
	attribute pinnum of TST_SDA: signal is "C4";

	attribute pinnum of SW: signal is "N17,H18,L14,L13";

	attribute pinnum of BTN: signal is "V4";
	attribute pinnum of BTS: signal is "K17";
	attribute pinnum of BTW: signal is "D18";
	attribute pinnum of BTE: signal is "H13";
	attribute pinnum of BTC: signal is "V16";

	attribute pinnum of ROTA: signal is "K18";
	attribute pinnum of ROTB: signal is "G18";

end aboard;

architecture Behavioral of aboard is

-- Clock management
--------------------------------------------
	signal CLK: std_logic;	-- Buffered ETH_TX_CLK (25 MHz)
	signal SCK: std_logic;	-- 50 MHz derived from ETH_TX_CLK
--------------------------------------------

-- System wake up
--------------------------------------------
	signal PLL_REG: std_logic_vector(3 downto 0);
	signal PLL_DEL: std_logic_vector(3 downto 0);
	signal RX_SOF : std_logic;								-- RX (receive) frame
	signal DV_DEL : std_logic;
--------------------------------------------

-- Ethernet input package
--------------------------------------------
	signal RXB_CLK: std_logic;
	signal RXB_CNT: natural range 0 to 2047;
	signal RXB_DAT: std_logic_vector(7 downto 0);	-- merging input nibbles

	signal MAC_CHK: std_logic_vector(7 downto 0);
	signal UNICAST: std_logic;								-- MAC match, chip select

	signal FE_ADR:  std_logic_vector(47 downto 0);

--		FE_ADR(47 downto 44) switch
--		FE_ADR(43 downto 40) unicast
--		FE_ADR(39 downto 0)  packet ID

	signal HOST_ADR: std_logic_vector(47 downto 0);
--------------------------------------------

-- Command flags for a single-loop processor
--------------------------------------------
	signal CTR_REG: std_logic_vector(2 downto 0);
	signal SPI_REG: std_logic_vector(2 downto 0);
	signal THR_REG: std_logic_vector(15 downto 0);

	signal DO_THR: std_logic;
	signal DO_SPI: std_logic;
	signal DO_ADC: std_logic;

	signal Peak0: std_logic;
	signal Peak1: std_logic;

	signal SIGN: std_logic;
	signal TRIG: std_logic;
	
	signal TRG_LAT: std_logic;
	signal TRG_DEL: std_logic;
--------------------------------------------

-- ADC cycle
--------------------------------------------
	signal ADC_GO: std_logic;
	signal BIT_CNT: natural range 0 to 63;

	signal ADC0: std_logic_vector(13 downto 0);
	signal ADC1: std_logic_vector(13 downto 0);

	signal ADC_RG0: std_logic_vector(15 downto 0);
	signal ADC_RG1: std_logic_vector(15 downto 0);
--------------------------------------------

-- Readout cycle
--------------------------------------------
	signal LOG_GTE: std_logic;
	signal REG_CLK: std_logic;

	signal LOG_CNT: natural range 0 to 15;
	signal REG_CNT: natural range 0 to 63;
--------------------------------------------

-- Data storage registers
--------------------------------------------
	signal S0A0_MSB: std_logic_vector(7 downto 0);
	signal S0A0_LSB: std_logic_vector(7 downto 0);
	signal S0A1_MSB: std_logic_vector(7 downto 0);
	signal S0A1_LSB: std_logic_vector(7 downto 0);

	signal S1A0_MSB: std_logic_vector(7 downto 0);
	signal S1A0_LSB: std_logic_vector(7 downto 0);
	signal S1A1_MSB: std_logic_vector(7 downto 0);
	signal S1A1_LSB: std_logic_vector(7 downto 0);

	signal S2A0_MSB: std_logic_vector(7 downto 0);
	signal S2A0_LSB: std_logic_vector(7 downto 0);
	signal S2A1_MSB: std_logic_vector(7 downto 0);
	signal S2A1_LSB: std_logic_vector(7 downto 0);

	signal S3A0_MSB: std_logic_vector(7 downto 0);
	signal S3A0_LSB: std_logic_vector(7 downto 0);
	signal S3A1_MSB: std_logic_vector(7 downto 0);
	signal S3A1_LSB: std_logic_vector(7 downto 0);

	signal S4A0_MSB: std_logic_vector(7 downto 0);
	signal S4A0_LSB: std_logic_vector(7 downto 0);
	signal S4A1_MSB: std_logic_vector(7 downto 0);
	signal S4A1_LSB: std_logic_vector(7 downto 0);

	signal S5A0_MSB: std_logic_vector(7 downto 0);
	signal S5A0_LSB: std_logic_vector(7 downto 0);
	signal S5A1_MSB: std_logic_vector(7 downto 0);
	signal S5A1_LSB: std_logic_vector(7 downto 0);

	signal S6A0_MSB: std_logic_vector(7 downto 0);
	signal S6A0_LSB: std_logic_vector(7 downto 0);
	signal S6A1_MSB: std_logic_vector(7 downto 0);
	signal S6A1_LSB: std_logic_vector(7 downto 0);

	signal S7A0_MSB: std_logic_vector(7 downto 0);
	signal S7A0_LSB: std_logic_vector(7 downto 0);
	signal S7A1_MSB: std_logic_vector(7 downto 0);
	signal S7A1_LSB: std_logic_vector(7 downto 0);

	signal S8A0_MSB: std_logic_vector(7 downto 0);
	signal S8A0_LSB: std_logic_vector(7 downto 0);
	signal S8A1_MSB: std_logic_vector(7 downto 0);
	signal S8A1_LSB: std_logic_vector(7 downto 0);

	signal S9A0_MSB: std_logic_vector(7 downto 0);
	signal S9A0_LSB: std_logic_vector(7 downto 0);
	signal S9A1_MSB: std_logic_vector(7 downto 0);
	signal S9A1_LSB: std_logic_vector(7 downto 0);
--------------------------------------------

-- Event number
--------------------------------------------
	signal EVT_CNT: natural range 0 to 2147483647;
	signal EVT_NUM: std_logic_vector(31 downto 0);
--------------------------------------------

-- Ethernet packet length
--------------------------------------------
--	constant DAT_LIM: natural:= 46;
	signal TXB_CNT: natural range 0 to 127;

	signal TX_CYCLE: std_logic;
--------------------------------------------

-- Checksum variables
--------------------------------------------
	signal CRC_CLK: std_logic;
	signal CRC_INI: std_logic;
	signal CRC_DAT: std_logic_vector(7 downto 0);
	signal CRC_CRC: std_logic_vector(31 downto 0);
	signal CRC_RES: std_logic_vector(31 downto 0);
--------------------------------------------

-- Beetle control
--------------------------------------------
	signal TST_DAT: std_logic;
	signal TST_DEN: std_logic;
--------------------------------------------


	signal TP_CNT: natural range 0 to 67108863;
	signal TPULSE: std_logic;

	signal TMP0: std_logic;
	signal TMP1: std_logic;

-- declaring PLL component
--------------------------------------------
component pll
	port (CLKIN_IN:   in std_logic;
			CLK0_OUT:  out std_logic;
			CLK2X_OUT: out std_logic);
end component;
--------------------------------------------

-- declaring CRC component
--------------------------------------------
	component CRC
		port (
			  Clock: in std_logic;
			  Init : in std_logic;
			  Data : in std_logic_vector(7 downto 0);
			  CRC  : out std_logic_vector(31 downto 0)
			  );
	end component;
--------------------------------------------

-- declaring comparator component
--------------------------------------------
	component COMP
		port (
				a: in std_logic_vector(15 downto 0);
				b: in std_logic_vector(15 downto 0);
				a_gt_b: out std_logic
				);
	end component;
--------------------------------------------

-- OK, here the entity starts


begin

-- Permanent assignments
--------------------------------------------

-- Resolve pin conflicts
--------------------------------------------
	AMP_SHDN <= '0';

	DAC_CLR  <= '0';
	DAC_CS   <= '1';
	DS_WIRE  <= 'Z';

-- M2..M0 are selected by jumpers
	FPGA_M0  <= 'Z';
	FPGA_M1  <= 'Z';
--	FPGA_M2  <= 'Z';
	FPGA_INIT_B <= '1';

	LCD_E    <= '0';
	LCD_RS   <= '0';
	LCD_RW   <= '0';

	PS2_CLK  <= '0';
	PS2_DATA <= '0';

	RS232_DCE_TXD <= '0';
	RS232_DTE_TXD <= '0';

	SD_A     <= (others => '0');
	SD_BA    <= "00";
	SD_CAS   <= '1'; -- NOP
	SD_CK_N  <= '1';
	SD_CK_P  <= '0';
	SD_CKE   <= '0'; -- disable clock
	SD_CS    <= '1'; -- disable command decoder
	SD_DQ    <= (others => 'Z');
	SD_LDM   <= '0';
	SD_LDQS  <= 'Z';
	SD_RAS   <= '1'; -- NOP
	SD_UDM   <= '0';
	SD_UDQS  <= 'Z';
	SD_WE    <= '1'; -- NOP

	SF_A1    <= (others => 'Z');
	SF_A0    <= (others => '0');
	SF_BYTE  <= '0';
	SF_CE0   <= '1';
	SF_D     <= (others => 'Z');
	SF_OE    <= '1';
	SF_WE    <= '1';

	SPI_SS_B <= '1';
	SPI_ALT_CS_JP11 <= 'Z';

	VGA_BLUE  <='0';
	VGA_GREEN <='0';
	VGA_HSYNC <='0';
	VGA_RED   <='0';
	VGA_VSYNC <='0';
--------------------------------------------

-- FE clock fanout
--------------------------------------------
	SYS_ADR <= (others => 'Z');
--	SYS_CLK <= (others => 'Z');

	SYS_CLK(0) <= TX_CYCLE;
	SYS_CLK(1) <= LOG_GTE;
	SYS_CLK(2) <= TRG_LAT;
	SYS_CLK(3) <= TPULSE;

--	SYS_TRG <= (others => 'Z');
	SYS_TRG(0) <= ADC_GO;
	SYS_TRG(1) <= TMP0;
	SYS_TRG(2) <= TMP1;
	SYS_TRG(3) <= '1';

--	SYS_RES <= (others => 'Z');
	SYS_RES(0) <= '1'; -- reference for AD8029
	SYS_RES(1) <= Peak0;
	SYS_RES(2) <= Peak1;
	SYS_RES(3) <= TRIG; 
	
	process(SCK, BIT_CNT, ADC0, ADC1, THR_REG)
	begin
			case BIT_CNT is
				when 4 =>
					TMP0 <= ADC0(13);
					TMP1 <= THR_REG(13);
				when 5 =>
					TMP0 <= ADC0(12);
					TMP1 <= THR_REG(12);
				when 6 =>
					TMP0 <= ADC0(11);
					TMP1 <= THR_REG(11);
				when 7 =>
					TMP0 <= ADC0(10);
					TMP1 <= THR_REG(10);
				when 8 =>
					TMP0 <= ADC0(9);
					TMP1 <= THR_REG(9);
				when 9 =>
					TMP0 <= ADC0(8);
					TMP1 <= THR_REG(8);
				when 10 =>
					TMP0 <= ADC0(7);
					TMP1 <= THR_REG(7);
				when 11 =>
					TMP0 <= ADC0(6);
					TMP1 <= THR_REG(6);
				when 12 =>
					TMP0 <= ADC0(5);
					TMP1 <= THR_REG(5);
				when 13 =>
					TMP0 <= ADC0(4);
					TMP1 <= THR_REG(4);
				when 14 =>
					TMP0 <= ADC0(3);
					TMP1 <= THR_REG(3);
				when 15 =>
					TMP0 <= ADC0(2);
					TMP1 <= THR_REG(2);
				when 16 =>
					TMP0 <= ADC0(1);
					TMP1 <= THR_REG(1);
				when 17 =>
					TMP0 <= ADC0(0);
					TMP1 <= THR_REG(0);
				when 20 =>
					TMP0 <= ADC1(13);
					TMP1 <= THR_REG(13);
				when 21 =>
					TMP0 <= ADC1(12);
					TMP1 <= THR_REG(12);
				when 22 =>
					TMP0 <= ADC1(11);
					TMP1 <= THR_REG(11);
				when 23 =>
					TMP0 <= ADC1(10);
					TMP1 <= THR_REG(10);
				when 24 =>
					TMP0 <= ADC1(9);
					TMP1 <= THR_REG(9);
				when 25 =>
					TMP0 <= ADC1(8);
					TMP1 <= THR_REG(8);
				when 26 =>
					TMP0 <= ADC1(7);
					TMP1 <= THR_REG(7);
				when 27 =>
					TMP0 <= ADC1(6);
					TMP1 <= THR_REG(6);
				when 28 =>
					TMP0 <= ADC1(5);
					TMP1 <= THR_REG(5);
				when 29 =>
					TMP0 <= ADC1(4);
					TMP1 <= THR_REG(4);
				when 30 =>
					TMP0 <= ADC1(3);
					TMP1 <= THR_REG(3);
				when 31 =>
					TMP0 <= ADC1(2);
					TMP1 <= THR_REG(2);
				when 32 =>
					TMP0 <= ADC1(1);
					TMP1 <= THR_REG(1);
				when 33 =>
					TMP0 <= ADC1(0);
					TMP1 <= THR_REG(0);
				when others =>
					TMP0 <= '0';
					TMP1 <= '0';
			end case;
	end process;
--------------------------------------------

-- Ethernet collision is not used
--------------------------------------------
	ETH_TX_ERR <= '0';
	ETH_MDIO   <= 'Z';
	ETH_MDC    <= '0';
--------------------------------------------

-- FE data bus is not used
--------------------------------------------
	USB_nTXE <= TPULSE; --'Z';
	USB_nRXF <= 'Z';
	USB_DAT  <= (others => 'Z');
--------------------------------------------

-- Frequency  multiplier
--------------------------------------------
U0: pll
		port map (
			CLKIN_IN     => ETH_TX_CLK,
			CLK0_OUT     => CLK,
			CLK2X_OUT    => SCK);
--------------------------------------------

-- !!! Wake up !!!
-- Capture the frame delimiter (Result CLK <-)
--------------------------------------------
	process(ETH_RX_CLK, ETH_RXD)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			PLL_DEL <= ETH_RXD;
		end if;
	end process;
--------------	
	process(ETH_RX_CLK, PLL_DEL)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '0') then
			PLL_REG <= PLL_DEL;
		end if;
	end process;
--------------------------------------------

-- Start of the RX frame (CLK ->)
--------------------------------------------
	process(ETH_RX_CLK, ETH_RX_DV, ETH_RXD, PLL_REG)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			if (ETH_RX_DV = '1') then
				if (ETH_RXD = x"D" and PLL_REG = x"5") then
					RX_SOF <= '1';
				else
					null;
				end if;
			else
				RX_SOF <= '0';
			end if;
		end if;
	end process;
--------------------------------------------

-- Signal for the clock synchronisation (CLK ->)
--------------------------------------------
	process(ETH_RX_CLK, ETH_RX_DV)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			DV_DEL <= ETH_RX_DV;
		end if;
	end process;
--------------------------------------------

-- RX nibble swapper and data clock (CLK <-)
-- (synchronize phase with RX_SOF signal)
--------------------------------------------
	process(ETH_RX_CLK, DV_DEL, RX_SOF)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '0') then
			if (DV_DEL = '1' and RX_SOF = '0') then
				RXB_CLK <= '0';
			else
				RXB_CLK <= not RXB_CLK;
			end if;
		end if;
	end process;
--------------------------------------------

-- RX byte counter (CLK ->)
--------------------------------------------
	process(RXB_CLK, RX_SOF)
	begin
		if (RXB_CLK'event and RXB_CLK = '1') then
			if (RX_SOF = '1') then
				RXB_CNT <= RXB_CNT + 1;
			else
				RXB_CNT <= 0;
			end if;
		end if;
	end process;
--------------------------------------------

-- Compile bytes from nibbles. Data are
-- ready for latching by (CLK ->) under
-- NIR_SWP = '0' condition
--------------------------------------------
	process(ETH_RX_CLK, RXB_CLK, ETH_RXD)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			if (RXB_CLK = '1') then
				RXB_DAT(3 downto 0) <= ETH_RXD;
			else
				null;
			end if;
		end if;
	end process;
--------------
	RXB_DAT(7 downto 4) <= ETH_RXD;
--------------------------------------------

-- Validate FE addressing (unicast)
--------------------------------------------
	process(ETH_RX_CLK, RXB_CNT, RXB_CLK, RXB_DAT)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			if (RXB_CNT = 1 and RXB_CLK = '0') then
				MAC_CHK <= RXB_DAT;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- Locally administered unicast
--------------------------------------------
	process(MAC_CHK)
	begin
		if (MAC_CHK(1 downto 0) = "10") then
			UNICAST <= '1';
		else
			UNICAST <= '0';
		end if;
	end process;
--------------------------------------------

-- Capture the RX data
--------------------------------------------
	process(ETH_RX_CLK, UNICAST, RXB_CLK, RXB_CNT,
			  MAC_CHK, RXB_DAT, DO_THR, DO_SPI)
	begin
		if (ETH_RX_CLK'event and ETH_RX_CLK = '1') then
			if (UNICAST = '1' and RXB_CLK = '0') then
				case RXB_CNT is
					when 2 =>
						FE_ADR(47 downto 40)   <= MAC_CHK;
						FE_ADR(39 downto 32)   <= RXB_DAT;
					when 3 =>
						FE_ADR(31 downto 24)   <= RXB_DAT;
					when 4 =>
						FE_ADR(23 downto 16)   <= RXB_DAT;
					when 5 =>
						FE_ADR(15 downto 8)    <= RXB_DAT;
					when 6 =>
						FE_ADR(7 downto 0)     <= RXB_DAT;
					when 7 =>
						HOST_ADR(47 downto 40) <= RXB_DAT;
					when 8 =>
						HOST_ADR(39 downto 32) <= RXB_DAT;
					when 9 =>
						HOST_ADR(31 downto 24) <= RXB_DAT;
					when 10 =>
						HOST_ADR(23 downto 16) <= RXB_DAT;
					when 11 =>
						HOST_ADR(15 downto 8)  <= RXB_DAT;
					when 12 =>
						HOST_ADR(7 downto 0)   <= RXB_DAT;

-- skip Dsize (2 bytes)
 
					when 15 =>
						   CTR_REG <= RXB_DAT(2 downto 0);
					when 16 =>
						if (DO_THR = '1') then
							THR_REG(15 downto 8) <= RXB_DAT;
						elsif (DO_SPI = '1') then
							SPI_REG <= RXB_DAT(2 downto 0);
						else
							null;
						end if;
					when 17 =>
						if (DO_THR = '1') then
							THR_REG(7 downto 0) <= RXB_DAT;
						else
							null;
						end if;
					when others =>
						null;
				end case;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- Hardware access
--------------------------------------------
	process(CTR_REG)
	begin
		case CTR_REG(1 downto 0) is
			when "01" =>
				DO_THR <= '1';
				DO_SPI <= '0';
				DO_ADC <= '0';				
			when "10" =>
				DO_THR <= '0';
				DO_SPI <= '1';
				DO_ADC <= '0';
			when "11" =>
				DO_THR <= '0';
				DO_SPI <= '0';
				DO_ADC <= '1';
			when others =>
				DO_THR <= '0';
				DO_SPI <= '0';
				DO_ADC <= '0';
		end case;
	end process;
--------------
	process(DO_SPI, SPI_REG, DO_ADC, SCK)
	begin
		if (DO_SPI = '1') then
			AMP_CS   <= SPI_REG(2);
			SPI_SCK  <= SPI_REG(1);
			SPI_MOSI <= SPI_REG(0);
		else
			AMP_CS   <= '1';
			SPI_SCK  <= DO_ADC and SCK;
			SPI_MOSI <= '0';
		end if;
	end process;
--------------
	SIGN <= CTR_REG(2);
--------------------------------------------

-- Bit counter @ 50 MHz
--------------------------------------------
	process(SCK, ADC_GO)
	begin
		if (SCK'event and SCK = '1') then
			if (ADC_GO = '1') then
				BIT_CNT <= 0;
			else
				BIT_CNT <= BIT_CNT + 1;
			end if;
		end if;
	end process;
--------------------------------------------

-- ADC sampling strobe
--------------------------------------------
	process(SCK, BIT_CNT)
	begin
		if (SCK'event and SCK = '0') then
			case BIT_CNT is
				when 0 =>
					ADC_GO <= '0';
				when 33 =>
					ADC_GO <= '1';
				when others =>
					null;
			end case;
		end if;
	end process;
--------------
	AD_CONV <= DO_ADC and ADC_GO;
--------------------------------------------

-- Compile data (-- Clock phase is inverted
-- for adjusment of system delays)
--------------------------------------------
	process(SCK, BIT_CNT, AMP_DOUT)
	begin
		if (SCK'event and SCK = '1') then
			case BIT_CNT is
				when 3 =>
					ADC0(13) <= not AMP_DOUT;
				when 4 =>
					ADC0(12) <= AMP_DOUT;
				when 5 =>
					ADC0(11) <= AMP_DOUT;
				when 6 =>
					ADC0(10) <= AMP_DOUT;
				when 7 =>
					ADC0(9) <= AMP_DOUT;
				when 8 =>
					ADC0(8) <= AMP_DOUT;
				when 9 =>
					ADC0(7) <= AMP_DOUT;
				when 10 =>
					ADC0(6) <= AMP_DOUT;
				when 11 =>
					ADC0(5) <= AMP_DOUT;
				when 12 =>
					ADC0(4) <= AMP_DOUT;
				when 13 =>
					ADC0(3) <= AMP_DOUT;
				when 14 =>
					ADC0(2) <= AMP_DOUT;
				when 15 =>
					ADC0(1) <= AMP_DOUT;
				when 16 =>
					ADC0(0) <= AMP_DOUT;
				when 19 =>
					ADC1(13) <= not AMP_DOUT;
				when 20 =>
					ADC1(12) <= AMP_DOUT;
				when 21 =>
					ADC1(11) <= AMP_DOUT;
				when 22 =>
					ADC1(10) <= AMP_DOUT;
				when 23 =>
					ADC1(9) <= AMP_DOUT;
				when 24 =>
					ADC1(8) <= AMP_DOUT;
				when 25 =>
					ADC1(7) <= AMP_DOUT;
				when 26 =>
					ADC1(6) <= AMP_DOUT;
				when 27 =>
					ADC1(5) <= AMP_DOUT;
				when 28 =>
					ADC1(4) <= AMP_DOUT;
				when 29 =>
					ADC1(3) <= AMP_DOUT;
				when 30 =>
					ADC1(2) <= AMP_DOUT;
				when 31 =>
					ADC1(1) <= AMP_DOUT;
				when 32 =>
					ADC1(0) <= AMP_DOUT;
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- Buffer ADC data
--------------------------------------------
	process(ADC_GO, ADC0, ADC1)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			ADC_RG0(15 downto 14) <= "00";
			ADC_RG1(15 downto 14) <= "00";

			ADC_RG0(13 downto 0) <= ADC0;
			ADC_RG1(13 downto 0) <= ADC1;
		end if;
	end process;
--------------

--------------------------------------------

-- Threshold (unsigned vectors)
--------------------------------------------
U1: COMP
		port map (
			a      => ADC_RG0,
			b      => THR_REG,
			a_gt_b => Peak0);
--------------
U2: COMP
		port map (
			a      => ADC_RG1,
			b      => THR_REG,
			a_gt_b => Peak1);
--------------------------------------------

-- Trigger condition
--------------------------------------------
	process(SIGN, Peak0, Peak1)
	begin
		if (SIGN = '1') then
			TRIG <= Peak0 or Peak1;
		else
			TRIG <= not (Peak0 and Peak1);
		end if;
	end process;
--------------------------------------------

-- Trigger pulse
--------------------------------------------
	process(ADC_GO, TRIG)
	begin
		if (ADC_GO'event and ADC_GO = '0') then
			TRG_LAT <= TRIG;
		end if;
	end process;
--------------------------------------------

-- Trigger latency
--------------------------------------------
	process(ADC_GO, TRG_LAT)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			TRG_DEL <= TRG_LAT;
		end if;
	end process;
--------------------------------------------

-- RAM write enable
--------------------------------------------
	process(ADC_GO, LOG_CNT, TRIG, TRG_DEL)
	begin
		if (ADC_GO'event and ADC_GO = '0') then
			if (LOG_CNT = 15) then
				LOG_GTE <= '0';				
			elsif ((TRIG = '1') and (TRG_DEL = '0')) then
				LOG_GTE <= '1';
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- Counter for number of samples
--------------------------------------------
	process(ADC_GO, LOG_GTE)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			if (LOG_GTE = '1') then
				LOG_CNT <= LOG_CNT + 1;
			else
				LOG_CNT <= 0;
			end if;
		end if;
	end process;
--------------------------------------------

-- Transmit gate
--------------------------------------------
	process(ADC_GO, LOG_CNT)
	begin
		if (ADC_GO'event and ADC_GO = '0') then
			case LOG_CNT is
				when 3 =>
					TX_CYCLE <= '1';
				when 12 =>
					TX_CYCLE <= '0';
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- Register write strobe
--------------------------------------------
	process(SCK, BIT_CNT)
	begin
		if (SCK'event and SCK = '0') then
			case BIT_CNT is
				when 4 =>
					REG_CLK <= '1';
				when 8 =>
					REG_CLK <= '0';
				when 12 =>
					REG_CLK <= '1';
				when 16 =>
					REG_CLK <= '0';
				when 20 =>
					REG_CLK <= '1';
				when 24 =>
					REG_CLK <= '0';
				when 28 =>
					REG_CLK <= '1';
				when 32 =>
					REG_CLK <= '0';
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- Register write address
--------------------------------------------
	process(REG_CLK, LOG_GTE)
	begin
		if (REG_CLK'event and REG_CLK = '1') then
			if (LOG_GTE = '1') then
				REG_CNT <= REG_CNT + 1;
			else
				REG_CNT <= 0;
			end if;
		end if;
	end process;
--------------------------------------------

-- Store event data
--------------------------------------------
	process(REG_CLK, REG_CNT, ADC_RG0, ADC_RG1)
	begin
		if (REG_CLK'event and REG_CLK = '0') then
			case REG_CNT is
				when 1 =>
					S0A0_MSB <= ADC_RG0(15 downto 8);
				when 2 =>
					S0A0_LSB <= ADC_RG0(7 downto 0);
				when 3 =>
					S0A1_MSB <= ADC_RG1(15 downto 8);
				when 4 =>
					S0A1_LSB <= ADC_RG1(7 downto 0);
--------------
				when 5 =>
					S1A0_MSB <= ADC_RG0(15 downto 8);
				when 6 =>
					S1A0_LSB <= ADC_RG0(7 downto 0);
				when 7 =>
					S1A1_MSB <= ADC_RG1(15 downto 8);
				when 8 =>
					S1A1_LSB <= ADC_RG1(7 downto 0);
--------------
				when 9 =>
					S2A0_MSB <= ADC_RG0(15 downto 8);
				when 10 =>
					S2A0_LSB <= ADC_RG0(7 downto 0);
				when 11 =>
					S2A1_MSB <= ADC_RG1(15 downto 8);
				when 12 =>
					S2A1_LSB <= ADC_RG1(7 downto 0);
--------------
				when 13 =>
					S3A0_MSB <= ADC_RG0(15 downto 8);
				when 14 =>
					S3A0_LSB <= ADC_RG0(7 downto 0);
				when 15 =>
					S3A1_MSB <= ADC_RG1(15 downto 8);
				when 16 =>
					S3A1_LSB <= ADC_RG1(7 downto 0);
--------------
				when 17 =>
					S4A0_MSB <= ADC_RG0(15 downto 8);
				when 18 =>
					S4A0_LSB <= ADC_RG0(7 downto 0);
				when 19 =>
					S4A1_MSB <= ADC_RG1(15 downto 8);
				when 20 =>
					S4A1_LSB <= ADC_RG1(7 downto 0);
--------------
				when 21 =>
					S5A0_MSB <= ADC_RG0(15 downto 8);
				when 22 =>
					S5A0_LSB <= ADC_RG0(7 downto 0);
				when 23 =>
					S5A1_MSB <= ADC_RG1(15 downto 8);
				when 24 =>
					S5A1_LSB <= ADC_RG1(7 downto 0);
--------------
				when 25 =>
					S6A0_MSB <= ADC_RG0(15 downto 8);
				when 26 =>
					S6A0_LSB <= ADC_RG0(7 downto 0);
				when 27 =>
					S6A1_MSB <= ADC_RG1(15 downto 8);
				when 28 =>
					S6A1_LSB <= ADC_RG1(7 downto 0);
--------------
				when 29 =>
					S7A0_MSB <= ADC_RG0(15 downto 8);
				when 30 =>
					S7A0_LSB <= ADC_RG0(7 downto 0);
				when 31 =>
					S7A1_MSB <= ADC_RG1(15 downto 8);
				when 32 =>
					S7A1_LSB <= ADC_RG1(7 downto 0);
--------------
				when 33 =>
					S8A0_MSB <= ADC_RG0(15 downto 8);
				when 34 =>
					S8A0_LSB <= ADC_RG0(7 downto 0);
				when 35 =>
					S8A1_MSB <= ADC_RG1(15 downto 8);
				when 36 =>
					S8A1_LSB <= ADC_RG1(7 downto 0);
--------------
				when 39 =>
					S9A0_MSB <= ADC_RG0(15 downto 8);
				when 40 =>
					S9A0_LSB <= ADC_RG0(7 downto 0);
				when 41 =>
					S9A1_MSB <= ADC_RG1(15 downto 8);
				when 42 =>
					S9A1_LSB <= ADC_RG1(7 downto 0);

				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- Ethernet packet number
--------------------------------------------
	process(LOG_GTE)
	begin
		if (LOG_GTE'event and LOG_GTE = '1') then
			EVT_CNT <= EVT_CNT + 1;
		end if;
	end process;
--------------
	EVT_NUM <= conv_std_logic_vector(EVT_CNT, 32);
--------------------------------------------

-- TX nibble swapper = CRC32 clock (CLK ->)
--------------------------------------------
	process(ETH_TX_CLK, TX_CYCLE)
	begin
		if (ETH_TX_CLK'event and ETH_TX_CLK = '1') then
			if (TX_CYCLE = '1') then
				CRC_CLK <= not CRC_CLK;
			else
				CRC_CLK <= '0';
			end if;
		end if;
	end process;
--------------------------------------------

-- TX byte counter (CLK ->) @ 12.5 MHz for
-- sending out 1 ethernet packet (1500 bytes)
-- per second for up to 375 Hz event trigger
-- rate (event size 4 bytes)
--------------------------------------------
	process(CRC_CLK, TX_CYCLE)
	begin
		if (CRC_CLK'event and CRC_CLK = '1') then
			if (TX_CYCLE = '1') then
				TXB_CNT <= TXB_CNT + 1;
			else
				TXB_CNT <= 0;
			end if;
		end if;
	end process;
--------------------------------------------

-- Drive ETH_TX_EN signal
-- Readout is postponed if it conflicts
-- with logging of zero-suppressed data 
--------------------------------------------
	process(ETH_TX_CLK, CRC_CLK, TXB_CNT)
	begin
		if (ETH_TX_CLK'event and ETH_TX_CLK = '0') then
			if (CRC_CLK = '1') then

-- Causes J/K/ = x55 for the receiver on the other end
				if (TXB_CNT = 1) then
					ETH_TX_EN <= '1';
				elsif (TXB_CNT = (22 + 46 + 4 + 1)) then
					ETH_TX_EN <= '0';
				else
					null;
				end if;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- Initialize CRC variables (CLK <-)
--------------------------------------------
	process(ETH_TX_CLK, TXB_CNT, CRC_CLK)
	begin
		if (ETH_TX_CLK'event and ETH_TX_CLK = '0') then
			if (TXB_CNT = 9) then
				if (CRC_CLK = '1') then
					CRC_INI <= '1';
				else
					CRC_INI <= '0';
				end if;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------

-- compile the Ethernet output packet
--------------------------------------------
	process(ETH_TX_CLK, TXB_CNT, CRC_CLK,
			  HOST_ADR, FE_ADR, EVT_NUM,
			  S0A0_MSB, S0A0_LSB, S0A1_MSB, S0A1_LSB,
			  S1A0_MSB, S1A0_LSB, S1A1_MSB, S1A1_LSB,
			  S2A0_MSB, S2A0_LSB, S2A1_MSB, S2A1_LSB,
			  S3A0_MSB, S3A0_LSB, S3A1_MSB, S3A1_LSB,
			  S4A0_MSB, S4A0_LSB, S4A1_MSB, S4A1_LSB,
			  S5A0_MSB, S5A0_LSB, S5A1_MSB, S5A1_LSB,
			  S6A0_MSB, S6A0_LSB, S6A1_MSB, S6A1_LSB,
			  S7A0_MSB, S7A0_LSB, S7A1_MSB, S7A1_LSB,
			  S8A0_MSB, S8A0_LSB, S8A1_MSB, S8A1_LSB,
			  S9A0_MSB, S9A0_LSB, S9A1_MSB, S9A1_LSB,
			  CRC_CRC, CRC_RES)
	begin
		if (ETH_TX_CLK'event and ETH_TX_CLK = '0') then

-- 6 times preamble 0x55
--------------------------------------------
			if ((1 <= TXB_CNT) and (TXB_CNT <= 7)) then
				ETH_TXD <= x"5";

-- Delimiter 0xD5
--------------------------------------------

			elsif (TXB_CNT = 8) then
				if (CRC_CLK = '1') then
					ETH_TXD <= x"5";
				else
					ETH_TXD <= x"D";
				end if;

-- Destination (Host) MAC address
--------------------------------------------
			elsif (TXB_CNT = 9) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(43 downto 40);
				else
					ETH_TXD <= HOST_ADR(47 downto 44);
				end if;
			elsif (TXB_CNT = 10) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(35 downto 32);
				else
					ETH_TXD <= HOST_ADR(39 downto 36);
				end if;
			elsif (TXB_CNT = 11) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(27 downto 24);
				else
					ETH_TXD <= HOST_ADR(31 downto 28);
				end if;
			elsif (TXB_CNT = 12) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(19 downto 16);
				else
					ETH_TXD <= HOST_ADR(23 downto 20);
				end if;
			elsif (TXB_CNT = 13) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(11 downto 8);
				else
					ETH_TXD <= HOST_ADR(15 downto 12);
				end if;
			elsif (TXB_CNT = 14) then
				if (CRC_CLK = '1') then
					ETH_TXD <= HOST_ADR(3 downto 0);
				else
					ETH_TXD <= HOST_ADR(7 downto 4);
				end if;

-- Source (Front-end) address
--------------------------------------------
			elsif (TXB_CNT = 15) then
				if (CRC_CLK = '1') then
					ETH_TXD <= FE_ADR(43 downto 40);
				else
					ETH_TXD <= FE_ADR(47 downto 44);
				end if;
			elsif (TXB_CNT = 16) then
				if (CRC_CLK = '1') then
					ETH_TXD <= FE_ADR(35 downto 32);
				else
					ETH_TXD <= FE_ADR(39 downto 36);
				end if;
			elsif (TXB_CNT = 17) then
				if (CRC_CLK = '1') then
					ETH_TXD <= EVT_NUM(27 downto 24);
				else
					ETH_TXD <= EVT_NUM(31 downto 28);
				end if;
			elsif (TXB_CNT = 18) then
				if (CRC_CLK = '1') then
					ETH_TXD <= EVT_NUM(19 downto 16);
				else
					ETH_TXD <= EVT_NUM(23 downto 20);
				end if;
			elsif (TXB_CNT = 19) then
				if (CRC_CLK = '1') then
					ETH_TXD <= EVT_NUM(11 downto 8);
				else
					ETH_TXD <= EVT_NUM(15 downto 12);
				end if;
			elsif (TXB_CNT = 20) then
				if (CRC_CLK = '1') then
					ETH_TXD <= EVT_NUM(3 downto 0);
				else
					ETH_TXD <= EVT_NUM(7 downto 4);
				end if;

-- Packet length is empty
--------------------------------------------

-- Data block
--------------------------------------------
			elsif (TXB_CNT = 23) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S0A0_MSB(3 downto 0);
				else
					ETH_TXD <= S0A0_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 24) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S0A0_LSB(3 downto 0);
				else
					ETH_TXD <= S0A0_LSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 25) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S0A1_MSB(3 downto 0);
				else
					ETH_TXD <= S0A1_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 26) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S0A1_LSB(3 downto 0);
				else
					ETH_TXD <= S0A1_LSB(7 downto 4);
				end if;
--------------
			elsif (TXB_CNT = 27) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S1A0_MSB(3 downto 0);
				else
					ETH_TXD <= S1A0_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 28) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S1A0_LSB(3 downto 0);
				else
					ETH_TXD <= S1A0_LSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 29) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S1A1_MSB(3 downto 0);
				else
					ETH_TXD <= S1A1_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 30) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S1A1_LSB(3 downto 0);
				else
					ETH_TXD <= S1A1_LSB(7 downto 4);
				end if;
--------------
			elsif (TXB_CNT = 31) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S2A0_MSB(3 downto 0);
				else
					ETH_TXD <= S2A0_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 32) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S2A0_LSB(3 downto 0);
				else
					ETH_TXD <= S2A0_LSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 33) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S2A1_MSB(3 downto 0);
				else
					ETH_TXD <= S2A1_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 34) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S2A1_LSB(3 downto 0);
				else
					ETH_TXD <= S2A1_LSB(7 downto 4);
				end if;
--------------
			elsif (TXB_CNT = 35) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S3A0_MSB(3 downto 0);
				else
					ETH_TXD <= S3A0_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 36) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S3A0_LSB(3 downto 0);
				else
					ETH_TXD <= S3A0_LSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 37) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S3A1_MSB(3 downto 0);
				else
					ETH_TXD <= S3A1_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 38) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S3A1_LSB(3 downto 0);
				else
					ETH_TXD <= S3A1_LSB(7 downto 4);
				end if;
--------------
			elsif (TXB_CNT = 39) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S4A0_MSB(3 downto 0);
				else
					ETH_TXD <= S4A0_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 40) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S4A0_LSB(3 downto 0);
				else
					ETH_TXD <= S4A0_LSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 41) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S4A1_MSB(3 downto 0);
				else
					ETH_TXD <= S4A1_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 42) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S4A1_LSB(3 downto 0);
				else
					ETH_TXD <= S4A1_LSB(7 downto 4);
				end if;
--------------
			elsif (TXB_CNT = 43) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S5A0_MSB(3 downto 0);
				else
					ETH_TXD <= S5A0_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 44) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S5A0_LSB(3 downto 0);
				else
					ETH_TXD <= S5A0_LSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 45) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S5A1_MSB(3 downto 0);
				else
					ETH_TXD <= S5A1_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 46) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S5A1_LSB(3 downto 0);
				else
					ETH_TXD <= S5A1_LSB(7 downto 4);
				end if;
--------------
			elsif (TXB_CNT = 47) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S6A0_MSB(3 downto 0);
				else
					ETH_TXD <= S6A0_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 48) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S6A0_LSB(3 downto 0);
				else
					ETH_TXD <= S6A0_LSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 49) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S6A1_MSB(3 downto 0);
				else
					ETH_TXD <= S6A1_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 50) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S6A1_LSB(3 downto 0);
				else
					ETH_TXD <= S6A1_LSB(7 downto 4);
				end if;
--------------
			elsif (TXB_CNT = 51) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S7A0_MSB(3 downto 0);
				else
					ETH_TXD <= S7A0_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 52) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S7A0_LSB(3 downto 0);
				else
					ETH_TXD <= S7A0_LSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 53) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S7A1_MSB(3 downto 0);
				else
					ETH_TXD <= S7A1_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 54) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S7A1_LSB(3 downto 0);
				else
					ETH_TXD <= S7A1_LSB(7 downto 4);
				end if;
--------------
			elsif (TXB_CNT = 55) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S8A0_MSB(3 downto 0);
				else
					ETH_TXD <= S8A0_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 56) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S8A0_LSB(3 downto 0);
				else
					ETH_TXD <= S8A0_LSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 57) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S8A1_MSB(3 downto 0);
				else
					ETH_TXD <= S8A1_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 58) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S8A1_LSB(3 downto 0);
				else
					ETH_TXD <= S8A1_LSB(7 downto 4);
				end if;
--------------
			elsif (TXB_CNT = 59) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S9A0_MSB(3 downto 0);
				else
					ETH_TXD <= S9A0_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 60) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S9A0_LSB(3 downto 0);
				else
					ETH_TXD <= S9A0_LSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 61) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S9A1_MSB(3 downto 0);
				else
					ETH_TXD <= S9A1_MSB(7 downto 4);
				end if;
			elsif (TXB_CNT = 62) then
				if (CRC_CLK = '1') then
					ETH_TXD <= S9A1_LSB(3 downto 0);
				else
					ETH_TXD <= S9A1_LSB(7 downto 4);
				end if;

-- CRC32
--------------------------------------------
			elsif (TXB_CNT = (22 + 46 + 1)) then
-- *** Important *** Use CRC_CRC Nibble here
				if (CRC_CLK = '1') then
					ETH_TXD <= CRC_CRC(3 downto 0);
				else
					ETH_TXD <= CRC_CRC(7 downto 4);
				end if;
			elsif (TXB_CNT = (22 + 46 + 2)) then
				if (CRC_CLK = '1') then
					ETH_TXD <= CRC_RES(11 downto 8);
				else
					ETH_TXD <= CRC_RES(15 downto 12);
				end if;
			elsif (TXB_CNT = (22 + 46 + 3)) then
				if (CRC_CLK = '1') then
					ETH_TXD <= CRC_RES(19 downto 16);
				else
					ETH_TXD <= CRC_RES(23 downto 20);
				end if;
			elsif (TXB_CNT = (22 + 46 + 4)) then
				if (CRC_CLK = '1') then
					ETH_TXD <= CRC_RES(27 downto 24);
				else
					ETH_TXD <= CRC_RES(31 downto 28);
				end if;

-- Packet length and remaining data 
--------------------------------------------
			else
				if (CRC_CLK = '1') then
					ETH_TXD <= (others => '0');
				else
					ETH_TXD <= (others => '0');
				end if;
			end if;
		end if;
	end process;
--------------------------------------------
--------------------------------------------


-- CRC data
--------------------------------------------
	process(TXB_CNT, HOST_ADR, FE_ADR, EVT_NUM,
			  S0A0_MSB, S0A0_LSB, S0A1_MSB, S0A1_LSB,
			  S1A0_MSB, S1A0_LSB, S1A1_MSB, S1A1_LSB,
			  S2A0_MSB, S2A0_LSB, S2A1_MSB, S2A1_LSB,
			  S3A0_MSB, S3A0_LSB, S3A1_MSB, S3A1_LSB,
			  S4A0_MSB, S4A0_LSB, S4A1_MSB, S4A1_LSB,
			  S5A0_MSB, S5A0_LSB, S5A1_MSB, S5A1_LSB,
			  S6A0_MSB, S6A0_LSB, S6A1_MSB, S6A1_LSB,
			  S7A0_MSB, S7A0_LSB, S7A1_MSB, S7A1_LSB,
			  S8A0_MSB, S8A0_LSB, S8A1_MSB, S8A1_LSB,
			  S9A0_MSB, S9A0_LSB, S9A1_MSB, S9A1_LSB)
	begin

-- Destination (Host) MAC address
--------------------------------------------
		if (TXB_CNT = 9) then
			CRC_DAT <= HOST_ADR(47 downto 40);
		elsif (TXB_CNT = 10) then
			CRC_DAT <= HOST_ADR(39 downto 32);
		elsif (TXB_CNT = 11) then
			CRC_DAT <= HOST_ADR(31 downto 24);
		elsif (TXB_CNT = 12) then
			CRC_DAT <= HOST_ADR(23 downto 16);
		elsif (TXB_CNT = 13) then
			CRC_DAT <= HOST_ADR(15 downto 8);
		elsif (TXB_CNT = 14) then
			CRC_DAT <= HOST_ADR(7 downto 0);

-- Source (Front-end) address
--------------------------------------------
		elsif (TXB_CNT = 15) then
			CRC_DAT <= FE_ADR(47 downto 40);
		elsif (TXB_CNT = 16) then
			CRC_DAT <= FE_ADR(39 downto 32);
		elsif (TXB_CNT = 17) then
			CRC_DAT <= EVT_NUM(31 downto 24);
		elsif (TXB_CNT = 18) then
			CRC_DAT <= EVT_NUM(23 downto 16);
		elsif (TXB_CNT = 19) then
			CRC_DAT <= EVT_NUM(15 downto 8);
		elsif (TXB_CNT = 20) then
			CRC_DAT <= EVT_NUM(7 downto 0);

-- Packet length is empty
--------------------------------------------

-- Data block
--------------------------------------------
		elsif (TXB_CNT = 23) then
			CRC_DAT <= S0A0_MSB;
		elsif (TXB_CNT = 24) then
			CRC_DAT <= S0A0_LSB;
		elsif (TXB_CNT = 25) then
			CRC_DAT <= S0A1_MSB;
		elsif (TXB_CNT = 26) then
			CRC_DAT <= S0A1_LSB;
--------------
		elsif (TXB_CNT = 27) then
			CRC_DAT <= S1A0_MSB;
		elsif (TXB_CNT = 28) then
			CRC_DAT <= S1A0_LSB;
		elsif (TXB_CNT = 29) then
			CRC_DAT <= S1A1_MSB;
		elsif (TXB_CNT = 30) then
			CRC_DAT <= S1A1_LSB;
--------------
		elsif (TXB_CNT = 31) then
			CRC_DAT <= S2A0_MSB;
		elsif (TXB_CNT = 32) then
			CRC_DAT <= S2A0_LSB;
		elsif (TXB_CNT = 33) then
			CRC_DAT <= S2A1_MSB;
		elsif (TXB_CNT = 34) then
			CRC_DAT <= S2A1_LSB;
--------------
		elsif (TXB_CNT = 35) then
			CRC_DAT <= S3A0_MSB;
		elsif (TXB_CNT = 36) then
			CRC_DAT <= S3A0_LSB;
		elsif (TXB_CNT = 37) then
			CRC_DAT <= S3A1_MSB;
		elsif (TXB_CNT = 38) then
			CRC_DAT <= S3A1_LSB;
--------------
		elsif (TXB_CNT = 39) then
			CRC_DAT <= S4A0_MSB;
		elsif (TXB_CNT = 40) then
			CRC_DAT <= S4A0_LSB;
		elsif (TXB_CNT = 41) then
			CRC_DAT <= S4A1_MSB;
		elsif (TXB_CNT = 42) then
			CRC_DAT <= S4A1_LSB;
--------------
		elsif (TXB_CNT = 43) then
			CRC_DAT <= S5A0_MSB;
		elsif (TXB_CNT = 44) then
			CRC_DAT <= S5A0_LSB;
		elsif (TXB_CNT = 45) then
			CRC_DAT <= S5A1_MSB;
		elsif (TXB_CNT = 46) then
			CRC_DAT <= S5A1_LSB;
--------------
		elsif (TXB_CNT = 47) then
			CRC_DAT <= S6A0_MSB;
		elsif (TXB_CNT = 48) then
			CRC_DAT <= S6A0_LSB;
		elsif (TXB_CNT = 49) then
			CRC_DAT <= S6A1_MSB;
		elsif (TXB_CNT = 50) then
			CRC_DAT <= S6A1_LSB;
--------------
		elsif (TXB_CNT = 51) then
			CRC_DAT <= S7A0_MSB;
		elsif (TXB_CNT = 52) then
			CRC_DAT <= S7A0_LSB;
		elsif (TXB_CNT = 53) then
			CRC_DAT <= S7A1_MSB;
		elsif (TXB_CNT = 54) then
			CRC_DAT <= S7A1_LSB;
--------------
		elsif (TXB_CNT = 55) then
			CRC_DAT <= S8A0_MSB;
		elsif (TXB_CNT = 56) then
			CRC_DAT <= S8A0_LSB;
		elsif (TXB_CNT = 57) then
			CRC_DAT <= S8A1_MSB;
		elsif (TXB_CNT = 58) then
			CRC_DAT <= S8A1_LSB;
--------------
		elsif (TXB_CNT = 59) then
			CRC_DAT <= S9A0_MSB;
		elsif (TXB_CNT = 60) then
			CRC_DAT <= S9A0_LSB;
		elsif (TXB_CNT = 61) then
			CRC_DAT <= S9A1_MSB;
		elsif (TXB_CNT = 62) then
			CRC_DAT <= S9A1_LSB;

-- Packet length and remaning data block
-------------------------------------------
		else
			CRC_DAT <= (others => '0');
		end if;
	end process;
--------------------------------------------
--------------------------------------------

-- Calculate CRC32 for transmitted data
-------------------------------------------
	U3:CRC
	port map (
		Clock => CRC_CLK,
		Init  =>	CRC_INI,
		Data  =>	CRC_DAT,
		CRC   =>	CRC_CRC);
--------------------------------------------

-- Keep the CRC result (the last data byte)
--------------------------------------------
	process(CRC_CLK, TXB_CNT, CRC_CRC)
	begin
		if (CRC_CLK'event and CRC_CLK = '0') then
			if (TXB_CNT = (22 + 46 + 1)) then
				CRC_RES <= CRC_CRC;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------
--------------------------------------------

-- Reserved for future use
--------------------------------------------
	TST_SYN <= '0';
	TST_TRG <= '0';
	TST_PUL <= '0';
	TST_RES <= '0';

	TST_SCL <= 'Z';
	TST_SDA <= 'Z';
--------------------------------------------

-- Test pulse generator
--------------------------------------------
	process(Clock)
	begin
		if (Clock'event and Clock = '1') then
			TP_CNT <= TP_CNT + 1;
		end if;
	end process;
--------------
	process(Clock, TP_CNT)
	begin
		if (Clock'event and Clock = '0') then
			case TP_CNT is
				when 0 =>
					TPULSE <= '1';
				when 200 =>
					TPULSE <= '0';
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

end Behavioral;
