library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.NUMERIC_STD.ALL;

entity ALOG is
    port (
                SCK:   in std_logic;
		CONV: out std_logic;
		SDO:   in std_logic;

		ADC_R0: out std_logic_vector(13 downto 0);
		ADC_R1: out std_logic_vector(13 downto 0);

		TRG_LEV: in std_logic_vector(13 downto 0);  

		RAM_CLK: out std_logic
	);
end ALOG;

architecture Behavioral of ALOG is

-- signals
--------------------------------------------
	signal BIT_CNT: natural range 0 to 63;
	signal ADC_GO:  std_logic;

	signal ADC0: std_logic_vector(13 downto 0);
	signal ADC1: std_logic_vector(13 downto 0);

	signal RAM_GTE: std_logic;
	signal RAM_STB: std_logic;

	signal CYC_CNT: natural range 0 to 16777215; -- ~1sec @ 12.5 MHz
	signal CYC_END: std_logic;
--------------------------------------------

--------------------------------------------

begin
--------------------------------------------

-- Permanent assignments
--------------------------------------------
	CONV <= ADC_GO;
--------------------------------------------

-- Bit counter
--------------------------------------------
	process(SCK, ADC_GO)
	begin
		if (SCK'event and SCK = '1') then
			if (ADC_GO = '1') then
				BIT_CNT <= 0;
			else
				BIT_CNT <= BIT_CNT + 1;
			end if;
		end if
	end process;
--------------------------------------------

-- ADC sampling strobe
--------------------------------------------
	process(SCK, BIT_CNT)
	begin
		if (SCK'event and SCK = '0') then
			case BIT_CNT is
				when 0 =>
					ADC_GO = '0';
				when 33 =>
					ADC_GO = '1';
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- Compile data
--------------------------------------------
	process(SCK, BIT_CNT, SDO)
	begin
		if (SCK'event and SCK = '0') then
			case BIT_CNT is
				when 2 =>
					ADC0(13) <= SDO;
				when 3 =>
					ADC0(12) <= SDO;
				when 4 =>
					ADC0(11) <= SDO;
				when 5 =>
					ADC0(10) <= SDO;
				when 6 =>
					ADC0(9) <= SDO;
				when 7 =>
					ADC0(8) <= SDO;
				when 8 =>
					ADC0(7) <= SDO;
				when 9 =>
					ADC0(6) <= SDO;
				when 10 =>
					ADC0(5) <= SDO;
				when 11 =>
					ADC0(4) <= SDO;
				when 12 =>
					ADC0(3) <= SDO;
				when 13 =>
					ADC0(2) <= SDO;
				when 14 =>
					ADC0(1) <= SDO;
				when 15 =>
					ADC0(0) <= SDO;
				when 18 =>
					ADC1(13) <= SDO;
				when 19 =>
					ADC1(12) <= SDO;
				when 20 =>
					ADC1(11) <= SDO;
				when 21 =>
					ADC1(10) <= SDO;
				when 22 =>
					ADC1(9) <= SDO;
				when 23 =>
					ADC1(8) <= SDO;
				when 24 =>
					ADC1(7) <= SDO;
				when 25 =>
					ADC1(6) <= SDO;
				when 26 =>
					ADC1(5) <= SDO;
				when 27 =>
					ADC1(4) <= SDO;
				when 28 =>
					ADC1(3) <= SDO;
				when 29 =>
					ADC1(2) <= SDO;
				when 30 =>
					ADC1(1) <= SDO;
				when 31 =>
					ADC1(0) <= SDO;
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------------------------

-- Buffer ADC data
--------------------------------------------
	process(ADC_GO, ADC0, ADC1)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			ADC_R0 <= ADC0;
			ADC_R1 <= ADC1;
		end if;
	end process;
--------------------------------------------

-- Zero suppression
--------------------------------------------
	process(ADC_GO, TRG_LEV, ADC0, ADC1)
	begin
		if (ADC_GO'event and ADC_GO = '1') then
			if ((TRG_LEV <= ADC0) or (TRG_LEV <= ADC1)) then
				RAM_GTE <= '1';
			else
				RAM_GTE <= '0';
			end if;
		end if;
	end process;
--------------------------------------------

-- RAM write strobe
--------------------------------------------
	process(SCK, BIT_CNT)
	begin
		if (SCK'event and SCK = '0') then
			case BIT_CNT is
				when 1 =>
					RAM_STB <= '1';
				when 5 =>
					RAM_STB <= '0';
				when 9 =>
					RAM_STB <= '1';
				when 13 =>
					RAM_STB <= '0';
				when 17 =>
					RAM_STB <= '1';
				when 21 =>
					RAM_STB <= '0';
				when 25 =>
					RAM_STB <= '1';
				when 29 =>
					RAM_STB <= '0';
				when others =>
					null;
			end case;
		end if;
	end process;
--------------------------
	RAM_CLK <= RAM_GTE and RAM_STB;
--------------------------------------------

-- Size of data sample
--------------------------------------------
	process(SYS_RES, ADC_GO, TRG_LEV, ADC0, ADC1)
	begin
		if (SYS_RES = '1') then
			DAT_CNT <= 0;
		elsif (ADC_GO'event and ADC_GO = '1') then
			if ((TRG_LEV <= ADC0) or (TRG_LEV <= ADC1)) then
				DAT_CNT <= DAT_CNT + 4;
			else
				null;
			end if;
		end if;
	end process;
--------------------------------------------





end Behavioral;
